/*********************************************************************************
 This file includes the implementation of following classes:
 1) CCThread -- the control tread defined above the POSIX thread 
 2) ControlStatus -- the class that manages the status of the software

 Also included:
 3) function "CtrlFunction" -- code to be executed by each CCThread
 4) global variables initialization
*********************************************************************************/



#include "ctrlcenter.h"
#include <assert.h>
#include "DStones.h"
#include "DSbuffer.h"
#include "rtp_handler.h"
//#include "useragentsip.h"
#include <qdatetime.h>


/*********************************************************************************
 1) CCThread Class
*********************************************************************************/

CCThread::CCThread(ControlStatus* Stat, void * (*Func)(void *), void* Par) {
	assert(Stat!=0);
	Status = Stat;
	assert(Func!=0);
	Function = Func;
	assert(Par!=0);
	Parameters = Par;
}

int CCThread::run() {
	return pthread_create( &THR, NULL, Function, Parameters);
}

void CCThread::join() {
	pthread_join( THR, NULL);
}

void CCThread::exit(int* RET) {
	pthread_exit(RET); //*RET must be the return value of THR, i.e. the return of pthread_create
}


/*********************************************************************************
 2) ControlStatus Class
*********************************************************************************/

ControlStatus::ControlStatus() {
	initialize(NULL);
	int DIM = 4;
	SessionList = new queue<SessionPar>(DIM);
	TableList = new queue<Record>(-1);
	busy = false;
	mode = 1;
	ringback = new bool[DIM];
	ringtone = new bool[DIM];
	busytone = new bool[DIM];
	failtone = new bool[DIM];
	for (int k=0; k<DIM; k++) {
		ringback[k] = false;
		ringtone[k] = false;
		busytone[k] = false;
		failtone[k] = false;
	}
	Win = 0;
}

void ControlStatus::initialize(UserAgentWin *W) {
	active_audio=0;
	active_video=0;
	active_chat=0;
	busy=0;
	session_counter=0;
	Win = W;
}

void ControlStatus::set_mode(int M) {
	assert(M>=1 && M<=3);
	mode = M;
}

int ControlStatus::get_mode(){
	return mode;
}

bool ControlStatus::new_outgoing_req(SessionPar* SP) {
	//check whether bandwidth is available or not...
	//...
	assert(SP>=0);
	//for now, just return the inverse of the "busy" bool value
	(busy)?( append("  Outgoing call not possible")):(append("  Outgoing call initiated"));
	return !busy;
}

bool ControlStatus::new_incoming_req(SessionPar* SP){
	//check whether bandwidth is available or not...
	//...
	assert(SP>=0);
	//for now, just return the inverse of the "busy" bool value
	(busy)?( append("  Incoming call rejected")):( append("  Incoming call accepted"));
	return !busy;
}

int ControlStatus::add_session(SessionPar* SP){
	assert(SP!=0);
	SP->Locale = NULL;
	SP->Remote = NULL;
	SP->buffer_a2n = NULL;
	SP->buffer_n2a = NULL;

	cout << "\nAdding a new active section..." << endl; 
	bool out = false;
    if(SP->session_status != active) {
	if(SessionList->push(SP) == 0) {
		if(SessionList->get_len()-SessionList->get_npack()==0)
			busy = true;
		session_counter++;
		SP->id = session_counter;
		cout << "\nSession established" << endl;
	} else {
		SP->id = -1; // session to be aborted due to lack of software resurces
		cout << "\nSession List full..." << endl;
		cout << "\nSession aborted" << endl;
		exit(-1);
	}
    }
    else {
	if(start_media_session(SP))
		if(out = start_RTP_session(SP))
			out = true;
		else {
			cout << "\nNo media found..." << endl;
			stop_media_session(SP);
		}
	else {
		cout <<  "\nNetwork unavailable..." << endl; 
		}

	if(out) {
		cout << "\nMedia and Network connected..." << endl;
		if(SessionList->push(SP) == 0) {
			if(SessionList->get_len()-SessionList->get_npack()==0)
				busy = true;
			session_counter++;
			SP->id = session_counter;
			cout << "\nSession established" << endl;
			
		} else {
			SP->id = -1; // session tO be aborted due to lack of software resurces
			cout << "\nSession List full..." << endl;
			cout << "\nSession aborted" << endl;
			exit(-1);
		} 
		if(SP->id!=-1) {
			cout << "\nStart media OK" << endl;
			start_media(SP);
		}
		else {
			cout << "\nStart media failed: Session aborted" << endl;
		}
		
	} else {
		SP->id = -1; //session to be aborted due to media or network problems
	}
     }
	return SP->id; 
     
}

void ControlStatus::start_session(SessionPar* SP) {
	bool out = false;
	if(start_media_session(SP))
		if(start_RTP_session(SP))
			out=true;
		else {
			cout << "\nNo media found..." << endl;
			stop_media_session(SP);
		}
	else {
		cout <<  "\nNetwork unavailable..." << endl; 
		}

	if(out) {
		cout << "\nMedia and Network connected..." << endl;
			cout << "\nStart media OK" << endl;
			start_media(SP);
		}
	else {
		cout << "\nStart media failed: Session aborted" << endl;
		}

}


/*
int ControlStatus::add_session(SessionPar* SP){
	assert(SP!=0);
	cout << "Adding a new active section..." << endl; 
	bool out = false;
	SP->active = false;
	out = start_RTP_session(SP);
	if (out) 
		out = start_media_session(SP);
	if (!out) 
		stop_media_session(SP);
	else {
		if(SessionList->push(SP) == 0) {
			session_counter++;
			SP->id = session_counter;
			cout << "Session established" << endl;
			
		} else {
			SP->id = -1; // session ti be aborted due to lack of software resurces
			out = false;
			cout << "Session List full..." << endl;
			cout << "Session aborted" << endl;
		} 
	}
	if (out) {
		cout << "Start media OK" << endl;
		start_media(SP);
	} else {
		cout << "Start media failed: Session aborted" << endl;
		stop_media_session(SP);
	}
	return SP->id; 
}
*/
bool ControlStatus::remove_session(int ID) {
	cout << "\nVDB --> removing session" << endl; 
	bool FLAG = false;
	QTime TimeHandler;
	TimeHandler.start();
	for(int i=0; i< (SessionList->get_npack()); i++){
		if((*SessionList)[i]->id==ID)
		{
			cout << "\nVDB -->   a) removing RTP, ID = " << ID << endl;
			stop_RTP_session((*SessionList)[i]);
			cout << "\nVDB -->   b) removing audio, ID = " << ID << endl;
			stop_media_session((*SessionList)[i]);
			cout << "\nVDB --> all is over"  << endl;
			int duration = - (TimeHandler.secsTo((*SessionList)[i]->start_time));
			SoftState.append((string)"  Line " + Data((*SessionList)[i]->pos_id).c_str() + (string)": Call Duration: " + Data(duration).c_str() + (string)"  sec");
			SessionPar* SP = (*SessionList)[i];
			if(SP->mCis != NULL) {
			//	(*SP->mCis)->end();
				delete SP->mCis;
				SP->mCis = NULL;
			}
			if(SP->mSis != NULL) {
			//	(*SP->mSis)->end();
				delete SP->mSis;
				SP->mSis = NULL;
			}
			cout <<"\nVDB --> #SP = " << SessionList->get_npack() << endl;
			SessionList->remove(i);
			cout <<"\nVDB --> #SP = " << SessionList->get_npack() << endl;
			cout << "\nVDB --> session removed!" << endl;
			
			if(SessionList->get_len()-SessionList->get_npack()>0)
				busy = false;

			//assert(SP!=NULL);
			//delete SP;
			FLAG = true;
			break;
		}
	}
	if(!FLAG) cout <<" 404: Session not found"<<endl;
	return FLAG;
}

SessionPar* ControlStatus::get_session(int ID) {
	for(int i=0; i< (SessionList->get_npack()); i++){
		if((*SessionList)[i]->id==ID){
			return (*SessionList)[i];
		}
	}
	cout <<"\nVDB 404: Session not found"<<endl;
	return NULL;
}

SessionPar* ControlStatus::get_session(char* CID) {
	printf("\nVDB --> CID = %s\n", CID);
	int s=-23;
	bool res = get_SSID(CID, &s);
	printf("\nVDB --> SSID = %d Result= %d\n", s, res);
	if (res) 
		return get_session(s);
	else 
		return NULL;
}

void ControlStatus::start_ringback(int P){
	assert(P>0 && P<=SoftState.get_list()->get_len());
	ringback[P-1] = true;
	R.SSID = P;
	strcpy(R.CID,"ringback");
	CCThread ring_thread(this, SoundFunction, &R);
	ring_thread.run();
}

void ControlStatus::start_ringtone(int P){
	cout<<"\nVDB --> start ringtone"<< endl;
	assert(P>0 && P<=SoftState.get_list()->get_len());
	ringtone[P-1] = true;
	R.SSID = P;
	strcpy(R.CID,"ringtone");
	CCThread ring_thread(this, SoundFunction, &R);
	ring_thread.run();
}

void ControlStatus::start_busytone(int P){
	assert(P>0 && P<=SoftState.get_list()->get_len());
	busytone[P-1] = true;
	R.SSID = P;
	strcpy(R.CID,"busytone");
	CCThread ring_thread(this, SoundFunction, &R);
	ring_thread.run();
}

void ControlStatus::start_failtone(int P){
	assert(P>0 && P<=SoftState.get_list()->get_len());
	failtone[P-1] = true;
	R.SSID = P;
	strcpy(R.CID,"failtone");
	CCThread ring_thread(this, SoundFunction, &R);
	ring_thread.run();
}

void ControlStatus::stop_ringing(int P) { 
	assert(P>0 && P<=SoftState.get_list()->get_len());
	ringtone[P-1] = false;
	ringback[P-1] = false;
}

void ControlStatus::stop_busytone(int P) {
	assert(P>0 && P<=SoftState.get_list()->get_len());
	busytone[P-1] = false;
}

void ControlStatus::stop_failtone(int P) {
	assert(P>0 && P<=SoftState.get_list()->get_len());
	failtone [P-1]= false;
}

bool ControlStatus::start_media(SessionPar* SP) {
	CCThread thread_R(&SoftState, RecFunction, &(SP->id));
	thread_R.run();
	CCThread thread_P(&SoftState, PlayFunction, &(SP->id));
	thread_P.run();
	return true;
}

bool ControlStatus::start_media_session(SessionPar* SP) {
	SP->buffer_n2a = new DSbuffer(&mutex2, 1780, (char*)"Buffer_N2A");
	SP->buffer_n2a->fill();
	SP->buffer_a2n = new DSbuffer(&mutex3, 1780, (char*)"Buffer_A2N");
	SP->buffer_a2n->fill();
	SP->Locale = new DSout();
	SP->Remote = new DSin();
	SP->Remote->set_parameters();
	SP->Locale->set_parameters();
	return true;
}


bool ControlStatus::start_media_session(int SessionID) {
	bool FLAG = false;
	for(int i=0; i<SessionList->get_npack(); i++){
		if((*SessionList)[i]->id==SessionID)
		{
			FLAG = start_media_session((*SessionList)[i]);
			break;
		}
	}
	return FLAG;
} 

bool ControlStatus::stop_media_session(SessionPar* SP) {
	assert(SP!=0);
	bool OUT =false;
	//SP->session_status = idle;
	//usleep(2500000);
//	cout << "\nVDB --> deleting buffer n2a" << endl;
	if(SP->buffer_n2a != NULL) delete SP->buffer_n2a;
//	cout << "\nVDB --> deleting Remote" << endl;
	if(SP->Remote != NULL)	delete SP->Remote;
//	cout << "\nVDB --> deleting buffer a2n" << endl;
	if(SP->buffer_a2n != NULL) delete SP->buffer_a2n;
//	cout << "\nVDB --> deleting Locale" << endl;
	if(SP->Locale != NULL) delete SP->Locale;
	cout << "\nVDB --> Media stopped" << endl;
	OUT = true;
	//media session stopped

	return OUT;
}

bool ControlStatus::stop_media_session(int SessionID) {
	bool OUT = false;
	bool FLAG = false;
	SessionPar* SP = NULL;
	for(int i=0; i<SessionList->get_npack(); i++){
		if((*SessionList)[i]->id==SessionID)
		{
			FLAG = true;
			SP = (*SessionList)[i];
			break;
		}
	}
	if(FLAG) {
		OUT = stop_media_session(SP);
	} else {
		OUT = false;
	} 
	return OUT;
}


bool ControlStatus::start_RTP_session(SessionPar* SP) {
	
	//obtain pointers to buffers and open the flow
	if(SP->RTP_SYM == NULL)
	{
	//printf();
		SP->RTP_SYM = new RTP_Sym(SP);
		RTP_Sym* rtp_sym = SP->RTP_SYM;
	
		//RTP_Sym* rtp_sym_rx = new RTP_Sym(SP);
		SP->ARX = new audio_rx(rtp_sym, SP->ptime, SP->buffer_n2a);
		SP->ARX->mute(SP->mute);
		cout << "gonna running audio_rx" << endl; 
		//SP->ARX->start();
		SP->ARX->detach();
		//SP->source_port +=2;
		//RTP_Sym* rtp_sym_tx = new RTP_Sym(SP);
		SP->ATX = new audio_tx(rtp_sym, SP->ptime, SP->buffer_a2n);
		SP->ATX->mute(SP->mute);
		cout << "gonna running audio_tx" << endl; 
		//SP->ATX->start();
		SP->ATX->detach();
	} 
	
	return true;
}

bool ControlStatus::start_RTP_session(int SessionID) {
	bool FLAG = false;
	for(int i=0; i<SessionList->get_npack(); i++){
		if((*SessionList)[i]->id==SessionID)
		{
			FLAG = start_RTP_session((*SessionList)[i]);
			break;
		}
	}
	return FLAG;
} 

bool ControlStatus::stop_RTP_session(SessionPar* SP) {
	cout << "\nVDB --> stopping RTP for SP = " << SP <<endl;
	assert(SP!=0);
	bool OUT =false;
	SP->session_status = idle;
	
	if(SP->RTP_SYM != NULL) {
		cout << "\nVDB --> stopping action" <<endl;
		SP->ATX->stop_action();
		SP->ARX->stop_action();
		
		cout << "\nVDB --> going to while" <<endl;
		bool cond = true;

		while(cond) {
			if(SP->ARX != NULL)
				if(!(SP->ARX->isRunning())) {
					printf("\nVDB --> deleting ARX\n");
					//delete SP->ARX;
					SP->ARX = NULL;
					printf("\nVDB --> ARX deleted\n");
				}
			if(SP->ATX != NULL)
				if(!(SP->ATX->isRunning())) {
					printf("\nVDB --> deleting ATX\n");
					//delete SP->ATX;
					SP->ATX = NULL;
					printf("\nVDB --> ATX deleted\n");
				}
			if((SP->ARX == NULL) && (SP->ATX == NULL)) 
				cond = false;
		}
	
		printf("\nVDB --> out from the while\n");
//		pthread_mutex_lock(&mutex1);
//		delete SP->ARX;
//		pthread_mutex_unlock(&mutex1);

//		SP->ARX = NULL;
//		delete SP->ATX;
//		SP->ATX = NULL;
		
		//delete SP->RTP_SYM;
		SP->RTP_SYM = NULL;
		//SP->ARX->close_socket();


		printf("\nVDB --> RTP stopped\n");
	}
	else {
		cout << "\nVDB --> nothing to do!"<<endl;
	}
	//RTP session stopped
	OUT = true;
	
	return OUT;
}


bool ControlStatus::stop_RTP_session(int SessionID) {
	//Stopping an RTP session...	
	bool OUT = false;
	bool FLAG = false;
	SessionPar* SP = NULL; 
	for(int i=0; i<SessionList->get_npack(); i++){
		if((*SessionList)[i]->id==SessionID)
		{
			FLAG = true;
			SP = (*SessionList)[i];
			break;
		}
	}
	if(FLAG) {
		OUT = stop_RTP_session(SP);
	} else {
		//Error: RTP session not found
		OUT = false;
	} 
	return OUT;
}

bool ControlStatus::get_SSID(char* cid, int* ssid) {
	printf("\nVDB --> CID = %s\n", cid);

	for(int i=0; i<TableList->get_npack(); i++){
		if(!strcmp((*TableList)[i]->CID, cid))
		{
			*ssid=(*TableList)[i]->SSID;
			return true;
		}
	}
	return false;	
}

bool ControlStatus::get_CID(int ssid, char** cid, int* len){
	for(int i=0; i<TableList->get_npack(); i++){
		if((*TableList)[i]->SSID==ssid)
		{
			*len=sprintf(*cid,"%s", (*TableList)[i]->CID);
			if(*len>0)	
				return true;
			else 
				break;
		}
	}
	return false;
}

bool* ControlStatus::get_ringback(int p) {
	assert(p>0 && p<=SoftState.get_list()->get_len());
	return &ringback[p-1];
}

bool* ControlStatus::get_ringtone(int p) {
	assert(p>0 && p<=SoftState.get_list()->get_len());
	return &ringtone[p-1];
}

bool* ControlStatus::get_busytone(int p) {
	assert(p>0 && p<=SoftState.get_list()->get_len());
	return &busytone[p-1];
}

bool* ControlStatus::get_failtone(int p) {
	assert(p>0 && p<=SoftState.get_list()->get_len());
	return &failtone[p-1];
}

double ControlStatus::get_neat_rx_bytes() {
	return audio_rx::get_neat_rx_bytes();
}

double ControlStatus::get_gross_rx_bytes() {
	return audio_rx::get_gross_rx_bytes();
}

double ControlStatus::get_neat_tx_bytes() {
	return audio_tx::get_neat_tx_bytes();
}

double ControlStatus::get_gross_tx_bytes() {
	return audio_tx::get_gross_tx_bytes();
}


void ControlStatus::set_rx_bytes(double Neat, double Gross) {
	audio_rx::get_neat_rx_bytes()=Neat;
	audio_rx::get_gross_rx_bytes()=Gross;
}

void ControlStatus::set_tx_bytes(double Neat, double Gross) {
	audio_tx::get_neat_tx_bytes()=Neat;
	audio_tx::get_gross_tx_bytes()=Gross;
}


/*********************************************************************************
 3) definition of CtrlFunctions
*********************************************************************************/

void *CtrlFunction(void *ptr)
{
	char *message;
	message = (char *) ptr;
	
	usleep(2000000);

	SessionPar* SP = new SessionPar;
	SP->id=1;
	SP->session_status = idle;
	SP->media_type=0;
	SP->Remote=NULL;
	SP->Locale=NULL;
	SP->buffer_a2n=NULL;
	SP->buffer_n2a=NULL;
	SP->dest_IP= "10.163.57.48";
	strcpy(SP->codec_name,"Prova");
	SP->b=64000; //bandwidth in bps
	SP->ptime=10; //packet time in ms
	SP->bpf=1; //bytes per frame
	SP->fpp=1; //frames per packet
	usleep(1000000);
	
	SoftState.start_ringtone(1);
	usleep(1000000);
	SoftState.stop_ringing(1);
	usleep(3000000);
	
	// new data session
	SoftState.add_session(SP);
	usleep(1000000);
	
	return NULL;
}

void *SoundFunction(void *ptr)
{
	Record parameters;
	strcpy(parameters.CID,((Record *)ptr)->CID);
	parameters.SSID = ((Record *)ptr)->SSID;

	bool* aux = 0;
	string path;
	path = PACKAGE_SOUND_DIR;
	int N=10;
	if(!strcmp(parameters.CID,"ringback")) {
		aux = SoftState.get_ringback(parameters.SSID);
		path += "call.pcm";
		//path +="/dtmf/ZERO.pcm";
		//sprintf(path,"../../src/RingTones/call.pcm");
		SoftState.append("  line" + (string)Data(parameters.SSID).c_str() + " Ringback...");
		N=15;
	} else if (!strcmp(parameters.CID,"ringtone")) {
		aux = SoftState.get_ringtone(parameters.SSID);
		//sprintf(path,"../../src/RingTones/ring.pcm");
		path += "ring.pcm";
		SoftState.append("  line" + (string)Data(parameters.SSID).c_str() + " Ringing...");
		N=20;
	} else if (!strcmp(parameters.CID,"busytone")) {
		aux = SoftState.get_busytone(parameters.SSID);
		//sprintf(path,"../../src/RingTones/busy.pcm");
		path +="busy.pcm";
		SoftState.append("  line" + (string)Data(parameters.SSID).c_str() + ": The line is busy");
		N=5;
	} else if (!strcmp(parameters.CID,"failtone")) {
		aux = SoftState.get_failtone(parameters.SSID); 
		//sprintf(path,"../../src/RingTones/notfound.pcm");
		path +="notfound.pcm";
		//path +="busy.pcm";
		SoftState.append("  line" + (string)Data(parameters.SSID).c_str() + ":  Attempt failed...");
		N=3;
	} else {
		SoftState.append("  Error: not specified tone");
		cout << "VDB-->  Error: not specified tone"<<endl;
		exit(-1);
	}
	
	DStones Toner;
	DSbuffer buffer(&mutex3,16384,(char*)"tone");
	Toner.set_device();
	while(*aux && N>0) {
		N--;
		usleep(500000);
		if(parameters.SSID==SoftState.Win->get_line()){
			//Toner.set_device();
			Toner.play_tone(reinterpret_cast<unsigned char*> (buffer.get()),(char*) path.c_str());
		}
		if(*aux) usleep(500000);
	}
	
	if(*aux) {
		SoftState.Win->SlotHungup(parameters.SSID);
	}
	return NULL;
}


void *RecFunction(void *ptr)
{
	int* ID = (int*) ptr;
	SessionPar* SP = NULL;
	bool FLAG = false;
	for(int i =0; i< SoftState.get_list()->get_npack(); i++) {
		if((*(SoftState.get_list()))[i]->id == *ID) {
			SP = (*(SoftState.get_list()))[i];
			FLAG = true;
			break;
		}
	}
	if(FLAG) {
		unsigned char buff[160];
		size_t len = 160;
		while (SP->session_status == active) {
			len = SP->Remote->rec(buff);
			SP->buffer_a2n->write(buff,len);
			//usleep(19000);
		}
	}
	else {
		//cout << "SBAGLIATO!!! REC su sessione non esistente" << endl;
		exit(-1);
	}
	return NULL;
}

void *PlayFunction(void *ptr)
{
	int* ID = (int*) ptr;
	//printf("ID nella funzione: %d\n", *ID);
	SessionPar* SP = NULL;
	bool FLAG = false;
	for(int i =0; i< SoftState.get_list()->get_npack(); i++) {
		if((*(SoftState.get_list()))[i]->id == *ID) {
			SP = (*(SoftState.get_list()))[i];
			FLAG = true;
			break;
		}
	}
	if(FLAG) {
		//SP->Locale->loop_play((unsigned char*)SP->buffer_n2a->get());
		cout << "playing audio" << endl;		
		char buff[160];
		int len = 160; ///bytes da leggere a ogni ciclo
		int len_r;
		while (SP->session_status == active) {
			len_r = SP->buffer_n2a->fetch(len, buff);
			if(len_r > 0) {
				//cout << "data to play: " << len_r << endl;
				SP->Locale->play((unsigned char*)buff);
			}
			else usleep(1000);
			usleep(1800);
		}
		//cout << "stop playing audio" <<endl;
	}
	else {
		//cout << "SBAGLIATO!!! PLAY su sessione non esistente" << endl;
		exit(-1);
	}
	return NULL;
}


/*********************************************************************************
 4) global variables initialization
*********************************************************************************/

//int counter = 1;
pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t mutex2 = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t mutex3 = PTHREAD_MUTEX_INITIALIZER;
ControlStatus SoftState;
//SoftState.initialize();



/*********************************************************************************
 End
*********************************************************************************/

