/***************************************************************************
 *   Copyright (C) 2005 by ,,,   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "ctrlcenter.h"
#include "rtp_handler.h"

#define AUDIO_SAMPLE_RATE	8000
#define AUDIO_SAMPLE_SIZE	16

// Size of jitter buffer in ms
// The jitter buffer is used to smooth playing out incoming RTP packets.
// The size of the buffer is also used as the expiry time in the ccRTP
// stack. Packets that have timestamp that is older than then size of
// the jitter buffer will not be sent out anymore.
#define JITTER_BUF_MS		80

const StaticPayloadType PAYLOADFMT = sptPCMU;

// Timing constants
const int SCHEDULINGTOUT = 40000;
const int EXPIRETOUT = 100000;
// This is irrelevant. 
//const int TIMESTAMP_RATE = 90000;

double audio_rx::neat_rx_bytes = 0.0;
double audio_rx::gross_rx_bytes = 0.0;
double audio_tx::neat_tx_bytes = 0.0;
double audio_tx::gross_tx_bytes = 0.0;

/// class RTP_Sym
RTP_Sym::RTP_Sym(SessionPar* SP) : SymmetricRTPSession(SP->source_IP ,SP->source_port) {
	
	//local_ip = LOCALHOST;
	//remote_ip = REMOTEHOST;
	//remote_ip = PROXY;
	local_ip = SP->source_IP;
	remote_ip = SP->dest_IP;	
	// Is that correct?
	if( ! local_ip ){  
		// this is equivalent to `! local_ip.isInetAddress()'
		cerr << "Sym: IP address is not correct!" << endl;
		std::exit(1);
	}
	ssrc = getLocalSSRC();
	setSchedulingTimeout(SP->ptime * 1000);
	setExpireTimeout(EXPIRETOUT);
//	UDPTransmit::setTypeOfService(SOCKET_IPTOS_LOWDELAY);
	setPayloadFormat(StaticPayloadFormat(PAYLOADFMT));
	addDestination( remote_ip, SP->dest_port );
	startRunning();

	if( isActive() )
            cout << "conection is active." << endl;
        else
            cerr << "conection is not active." << endl;
	
	
}


RTP_Sym::~RTP_Sym()
{
}


uint32 
RTP_Sym::getLastTimestamp(const SyncSource *src) const {
	if ( src && !isMine(*src) ) return 0L;
	
	recvLock.readLock();

	uint32 ts = 0;	
	if (src != NULL) {
		SyncSourceLink* srcm = getLink(*src);
		IncomingRTPPktLink* l = srcm->getFirst();
		
		while (l) {
			ts = l->getTimestamp();
			l = l->getSrcNext();
		}
	} else {
		IncomingRTPPktLink* l = recvFirst;
		
		while (l) {
			ts = l->getTimestamp();
			l = l->getNext();
		}
	}
	
	recvLock.unlock();
	return ts;
}

/// Class for audio receive
audio_rx::audio_rx( RTP_Sym* _rtp_session, unsigned short _ptime, DSbuffer* B )
{	
	buffer_rx = B;
	rtp_session=_rtp_session;
	ptime=_ptime;
	active = false;
	//Set parameters for output audio device
	//out.set_parameters();
	payload_size = AUDIO_SAMPLE_RATE/1000 * ptime;
	mute_flag = true;
}


audio_rx::~audio_rx()
{

}

void audio_rx::mute(bool M) {
	mute_flag=M;
}

void audio_rx::dtmf_sent(char P, int D) {
	int s;
	if(D<0) 
		s=20;
	else if(D>1600) 
		s=1600;
	else 
		s=D;
	unsigned char payload[s];
	
	char path[256];
	sprintf(path,"%s",PACKAGE_SOUND_DIR);
	FILE* FP = NULL;
	switch (P) {
		case '0':
			strcat(path,"/dtmf/ZERO.pcm");
			break;
		case '1':
			strcat(path,"/dtmf/ONE.pcm");
			break;
		case '2':
			strcat(path,"/dtmf/TWO.pcm");
			break;
		case '3':
			strcat(path,"/dtmf/THREE.pcm");
			break;
		case '4':
			strcat(path,"/dtmf/FOUR.pcm");
			break;
		case '5':
			strcat(path,"/dtmf/FIVE.pcm");
			break;
		case '6':
			strcat(path,"/dtmf/SIX.pcm");
			break;
		case '7':
			strcat(path,"/dtmf/SEVEN.pcm");
			break;
		case '8':
			strcat(path,"/dtmf/EIGHT.pcm");
			break;
		case '9':
			strcat(path,"/dtmf/NINE.pcm");
			break;
		case 'A':
			strcat(path,"/dtmf/A.pcm");
			break;
		case 'B':
			strcat(path,"/dtmf/B.pcm");
			break;
		case 'C':
			strcat(path,"/dtmf/C.pcm");
			break;
		case 'D':
			strcat(path,"/dtmf/D.pcm");
			break;
		case '*':
			strcat(path,"/dtmf/STAR.pcm");
			break;
		case '#':
			strcat(path,"/dtmf/SHARP.pcm");
			break;
		default:
			cout << "\nVDB --> Wrong DTMF" << endl; 
	}
	FP=fopen(path,"r");
	if(FP!=NULL) {
		cout << "\nVDB --> FP!=NULL " << path<< endl;
		fread(payload,1,s,FP);
		buffer_rx->write(payload, s);
		fclose(FP);
	}
}

void 
audio_rx::run(void) {

	const AppDataUnit* adu;
	unsigned long rtp_timestamp;
	unsigned int dtmf_pause = 0;
	unsigned short recvd_ptime = 10;
	active=true;
	bool pause = false;
	//TimerPort::setTimer(ptime);
	unsigned char* void_sample = new unsigned char[payload_size];
	for(int k=0; k<payload_size; k++) 
			void_sample[k]=0xff; //dummy[k]=char(k);
	while (active) {
		TimerPort::setTimer(ptime);
		adu = NULL;
		rtp_timestamp =rtp_session->getFirstTimestamp();
		pthread_mutex_lock(&mutex1);
		adu = rtp_session->getData(rtp_session->getFirstTimestamp());
		pthread_mutex_unlock(&mutex1);
		if (adu == NULL || adu->getSize() <= 0) {
			// There is no packet available. This may have
			// several reasons:
			// - the thread scheduling granularity does
			//   not match ptime
			// - packet lost
			// - packet delayed
			// Wait another cycle for a packet. The
			// jitter buffer will cope with this variation.
			if (adu) {
				delete adu;
				adu = NULL;
			} 
		}
		if(adu == NULL || (adu->getSize() <= 0)) {
			continue;
		}
		else{
			unsigned char *payload = const_cast<uint8 *>(adu->getData());
			payload_size = adu->getSize();
			neat_rx_bytes+=payload_size;
			gross_rx_bytes+=(payload_size+40);
			if(!mute_flag || adu->getType() == 101) {
				if(adu->getType() == 0 ) {
					// Determine received ptime
					recvd_ptime =(payload_size/(AUDIO_SAMPLE_RATE / 1000));
					buffer_rx->write(payload, payload_size);
				
					if (pause) {
							cout << "DTMF Pause (ms):" << dtmf_pause << endl <<endl;;
							pause = false;
					}
				}
				/// DTMF session
				if(adu->getType() == 101 ) {
					int digit;
					bool end;
					RtpDtmfPayload *rtpev = new RtpDtmfPayload;
					if(payload_size == sizeof(RtpDtmfPayload)){
						rtpev = (RtpDtmfPayload *)payload;
						digit = (unsigned int)rtpev->event;
						end = (unsigned int)rtpev->end; 
						rtpev->duration = ((unsigned short)payload[2])*256 + (unsigned short)payload[3];
						buffer_rx->write(void_sample, payload_size);
						if (rtpev->end){
							if(!pause){
								switch(digit){
								case DTMFdigit0:
									SoftState.append("  received DTMF event 0");
									cout << "RTP DTMF EVENT: 0" << endl;
									break;
								case DTMFdigit1:
									SoftState.append("  received DTMF event 1");
									cout << "RTP DTMF EVENT: 1" << endl;
									break;
								case DTMFdigit2:
									SoftState.append("  received DTMF event 2");
									cout << "RTP DTMF EVENT: 2" << endl;
									break;
								case DTMFdigit3:
									SoftState.append("  received DTMF event 3");
									cout << "RTP DTMF EVENT: 3" << endl;
									break;
								case DTMFdigit4:
									SoftState.append("  received DTMF event 4");
									cout << "RTP DTMF EVENT: 4" << endl;
									break;
								case DTMFdigit5:
									SoftState.append("  received DTMF event 5");
									cout << "RTP DTMF EVENT: 5" << endl;
									break;
								case DTMFdigit6:
									SoftState.append("  received DTMF event 6");
									cout << "RTP DTMF EVENT: 6" << endl;
									break;
								case DTMFdigit7:
									SoftState.append("  received DTMF event 7");
									cout << "RTP DTMF EVENT: 7" << endl;
									break;
								case DTMFdigit8:
									SoftState.append("  received DTMF event 8");
									cout << "RTP DTMF EVENT: 8" << endl;
									break;
								case DTMFdigit9:
									SoftState.append("  received DTMF event 9");
									cout << "RTP DTMF EVENT: 9" << endl;
									break;
								case DTMFdigitA:
									SoftState.append("  received DTMF event A");
									cout << "RTP DTMF EVENT: A" << endl;
									break;
								case DTMFdigitB:
									SoftState.append("  received DTMF event B");
									cout << "RTP DTMF EVENT: B" << endl;
									break;
								case DTMFdigitC:
									SoftState.append("  received DTMF event C");
									cout << "RTP DTMF EVENT: C" << endl;
									break;
								case DTMFdigitD:
									SoftState.append("  received DTMF event D");
									cout << "RTP DTMF EVENT: D" << endl;
									break;
								case DTMFdigitstar:
									SoftState.append("  received DTMF event *");
									cout << "RTP DTMF EVENT: *" << endl;
									break;
								case DTMFdigitsharp:
									SoftState.append("  received DTMF event #");
									cout << "RTP DTMF EVENT: #" << endl;
									break;
								default: ;
								}
								if(!mute_flag) {
									cout << "RTP DTMF END: "
									<< (unsigned int)rtpev->end << endl;
									cout << "RTP DTMF VOLUME: "
									<< (unsigned int)rtpev->volume << endl;	
									cout << "RTP DTMF DURATION: "
									<< (unsigned int)rtpev->duration  << endl;
									cout << "DTMF Duration (ms):"
									<< (unsigned int)rtpev->duration/(AUDIO_SAMPLE_RATE/1000) << endl;
									dtmf_pause = 0;
									// riprodurre tono
									pause = true;
								}
							}
							else
								dtmf_pause += recvd_ptime;
							
						}
					}
				}
				delete adu;
			}
			else {
				buffer_rx->write(void_sample, AUDIO_SAMPLE_RATE/1000 * ptime);
			}
		}
	}
	delete[] void_sample;
}

void audio_rx::stop_action() {
	active = false;
}

void audio_rx::close_socket() {
	terminate();
}

/// Class for audio transmit
audio_tx::audio_tx( RTP_Sym* _rtp_session, unsigned short _ptime, DSbuffer* B )
{	
	buffer_tx = B;
	rtp_session=_rtp_session;
	ptime=_ptime;
	active = false;
	//Set parameters for output audio device
	nsamples = (AUDIO_SAMPLE_RATE/1000 * ptime);
	payload_size = AUDIO_SAMPLE_RATE/1000 * ptime;
	payload = new unsigned char[payload_size];
	dtmf_payload_size = sizeof(RtpDtmfPayload);
	dtmf_payload = new RtpDtmfPayload;
	dtmf_send = false;
	n_pause = 0;
	n_dtmf = 0;
	mute_flag = true;
	event_payload.event = 0;
	event_payload.volume = 0;
	event_payload.reserved = 0;
	event_payload.end = true;
	event_payload.duration = 0;
}


audio_tx::~audio_tx()
{
	delete payload;
	delete dtmf_payload;
}

void audio_tx::mute(bool M) {
	mute_flag=M;
}

bool audio_tx::send_dtmf(RtpDtmfEvent* event) {
	bool res;
	if(dtmf_send == false) {
		// convert the event duration from timestamp unit to number of packets
		dtmf_send = true;
		n_dtmf = event->duration/ptime; 
		n_pause = event->pause/ptime;
	
		event_payload.event = char(event->digit);
		event_payload.volume = event->volume;
		event_payload.reserved = false;
		event_payload.end = false;
		event_payload.duration = 0;
		res = true;
	}
	else 
		res = false;
	delete event;
	return res;
}

/*
void audio_tx::run(void) {

//	int status;
//	struct timespec sleeptimer;
//	struct timeval debug_timer;
	active = true;
	// Synchronize the timestamp driven by the sampling rate
	// of the recording with the timestamp of the RTP session.
	// As the RTP session is already created in advance, the
	// RTP clock is a bit ahead already.
	//timestamp = rtp_session->getCurrentTimestamp() + nsamples;
	timestamp = rtp_session->getCurrentTimestamp();
	TimerPort::setTimer(ptime);
	timeval ttt;
	gettimeofday(&ttt,0);
//	double ref_time = ttt.tv_sec*(1.0e6)+ttt.tv_usec;

	int cnt = 0;
	unsigned char pt = 0;
	
	while (active) {		
		buffer_tx->fetch(payload_size, payload);
//		cout << dtmf_send << endl;
		// If timestamp is more than 1 payload size ahead of the clock of
		// the ccRTP stack, then drop the current payload and do not advance
		// the timestamp. This will happen if the DSP delivers more
		// sound samples than the set sample rate. To compensate for this
		// samples must be dropped.
		if(dtmf_send) {
			pt = 101;
			rtp_session->setMark(true);
			rtp_session->setPayloadFormat(DynamicPayloadFormat (pt, AUDIO_SAMPLE_RATE));
			if(event_payload.end == false) {
				cnt ++;
				event_payload.duration += nsamples;
				if(cnt == n_dtmf) {
					event_payload.end = true;
					cnt=0;
				}
			}
			else {
				cnt++;
				if(cnt >= n_pause) {
					dtmf_send = false;
					cnt = 0;
				}
			}
	
			pthread_mutex_lock(&mutex1);

			if(dtmf_send == false) {
				dtmf_send = true;
				timestamp -=  nsamples;
				*dtmf_payload = event_payload;
				dtmf_payload->end = false;
				dtmf_payload->duration = 0;		
				cout << "\nsending first dtmf: " << dtmf_payload_size << " bytes";
				rtp_session->putData(timestamp, (unsigned char*)dtmf_payload, dtmf_payload_size);
				rtp_session->setMark(false);
				for(int j=1; j<n_dtmf-1; j++) {
					cout << "\nsending a dtmf: " << (unsigned char*)dtmf_payload;
					rtp_session->putData(timestamp, (unsigned char*)dtmf_payload, dtmf_payload_size);
				}

				*dtmf_payload = event_payload;
				unsigned char temp = ((unsigned char*)dtmf_payload)[3];
				((unsigned char*)dtmf_payload)[3] = ((unsigned char*)dtmf_payload)[2];
				((unsigned char*)dtmf_payload)[2] = temp;

				cout << "\nsending last dtmf: " << (unsigned char*)dtmf_payload;
				rtp_session->putData(timestamp, (unsigned char*)dtmf_payload, dtmf_payload_size);
				for(int k=0; k<n_pause; k++) { 
					cout << "\nsending a pause:  " << (unsigned char*)dtmf_payload;
					rtp_session->putData(timestamp, (unsigned char*)dtmf_payload, dtmf_payload_size);
				}
				cout << endl;
				timestamp +=  ((n_dtmf + n_pause) * nsamples);
				dtmf_send = false;
			}	
			//rtp_session->putData(timestamp, (unsigned char*)dtmf_payload, dtmf_payload_size);
			pthread_mutex_unlock(&mutex1);
		} 
		else 
		{
			pt = 0;
			rtp_session->setPayloadFormat(StaticPayloadFormat((ost::StaticPayloadType) pt));

			pthread_mutex_lock(&mutex1);
			
			//auxtime=timevero; 
			rtp_session->putData(timestamp, payload, payload_size);
			pthread_mutex_unlock(&mutex1);
			timestamp += nsamples;
			if (timestamp <= rtp_session->getCurrentTimestamp() - (JITTER_BUF_MS / ptime) * nsamples){
				timestamp += nsamples * (JITTER_BUF_MS / ptime);
			}
		}

		// There is no data left in the DSP buffers to play out anymore.
		// So the timestamp must be in sync with the clock of the ccRTP
		// stack. It might get behind if the sound cards samples a bit
		// slower than the set sample rate. Advance the timestamp to get
		// in sync again.
		
//		cout << "\n tx : timer = " << TimerPort::getTimer() << " ... " ;
		Thread::sleep(TimerPort::getTimer());
//		cout << "\t " << TimerPort::getTimer();
//		cout << "  end = " << event_payload.end;
//		cout << " digit = "<< int(event_payload.event) << endl;
		TimerPort::incTimer(ptime);
			
	}

	is_running = false;
} 
*/
bool audio_tx::send_dtmf_in_band(char P, int D) {
	//unsigned char *payload = const_cast<uint8 *>(adu->getData());
	//int s=1120;
	int s;
	if(D<0) 
		s=20;
	else if(D>1600) 
		s=1600;
	else 
		s=D;
	
	unsigned char payload[s];
//	unsigned char	event_b : 8;
//	bit_wise_payload payload[s*8];
	
	char path[256];
	sprintf(path,"%s",PACKAGE_SOUND_DIR);
	FILE* FP = NULL;
	switch (P) {
		case '0':
			strcat(path,"/dtmf/ZERO.pcm");
			break;
		case '1':
			strcat(path,"/dtmf/ONE.pcm");
			break;
		case '2':
			strcat(path,"/dtmf/TWO.pcm");
			break;
		case '3':
			strcat(path,"/dtmf/THREE.pcm");
			break;
		case '4':
			strcat(path,"/dtmf/FOUR.pcm");
			break;
		case '5':
			strcat(path,"/dtmf/FIVE.pcm");
			break;
		case '6':
			strcat(path,"/dtmf/SIX.pcm");
			break;
		case '7':
			strcat(path,"/dtmf/SEVEN.pcm");
			break;
		case '8':
			strcat(path,"/dtmf/EIGHT.pcm");
			break;
		case '9':
			strcat(path,"/dtmf/NINE.pcm");
			break;
		case 'A':
			strcat(path,"/dtmf/A.pcm");
			break;
		case 'B':
			strcat(path,"/dtmf/B.pcm");
			break;
		case 'C':
			strcat(path,"/dtmf/C.pcm");
			break;
		case 'D':
			strcat(path,"/dtmf/D.pcm");
			break;
		case '*':
			strcat(path,"/dtmf/STAR.pcm");
			break;
		case '#':
			strcat(path,"/dtmf/SHARP.pcm");
			break;
		default:
			cout << "\nVDB --> Wrong DTMF" << endl; 
			return false;
	}
	FP=fopen(path,"r");
	if(FP!=NULL) {
		cout << "\nVDB --> FP!=NULL " << path<< endl;
		fread(payload,1,s,FP);
		buffer_tx->write(payload, s);
		fclose(FP);
		return false;
	}
	
	return true;
}

void audio_tx::run(void) {
	active = true;
	timestamp = rtp_session->getCurrentTimestamp();
	TimerPort::setTimer(ptime);
	timeval ttt;
	gettimeofday(&ttt,0);
	int cnt = -1;
	unsigned char pt = 0;
	unsigned char* dummy = new unsigned char[payload_size];
	for(int k=0; k<payload_size; k++) 
			dummy[k]=0xff; //dummy[k]=char(k);
	while (active) {		
		buffer_tx->fetch(payload_size, payload);
		if(!mute_flag){
			if(dtmf_send && cnt==-1)
				cnt = n_pause+n_dtmf;
			if(cnt>0) {
				//cout << "\nsend dtmf: cnt = " << cnt;
				pt = 101;
				rtp_session->setPayloadFormat(DynamicPayloadFormat (pt, AUDIO_SAMPLE_RATE));
				event_payload.duration += nsamples;
				if(cnt==(n_dtmf+n_pause)) {
					rtp_session->setMark(true);
					*dtmf_payload = event_payload;
					dtmf_payload->end = false;
					dtmf_payload->duration = 0;		
					//cout << "\nsending first dtmf: " << dtmf_payload_size << " bytes";
					pthread_mutex_lock(&mutex1);
					rtp_session->putData(timestamp, (unsigned char*)dtmf_payload, dtmf_payload_size);
					pthread_mutex_unlock(&mutex1);
					neat_tx_bytes+=dtmf_payload_size;
					gross_tx_bytes+=(dtmf_payload_size+40);
				}
				else if (cnt>(n_pause+1)) {
					rtp_session->setMark(false);
					//cout << "\nsending a dtmf";
					pthread_mutex_lock(&mutex1);				
					rtp_session->putData(timestamp, (unsigned char*)dtmf_payload, dtmf_payload_size);
					pthread_mutex_unlock(&mutex1);
					neat_tx_bytes+=dtmf_payload_size;
					gross_tx_bytes+=(dtmf_payload_size+40);
				}
				else if(cnt==(n_pause+1)) {
					*dtmf_payload = event_payload;
					unsigned char temp = ((unsigned char*)dtmf_payload)[3];
					((unsigned char*)dtmf_payload)[3] = ((unsigned char*)dtmf_payload)[2];
					((unsigned char*)dtmf_payload)[2] = temp;
					dtmf_payload->end = true;
					//cout << "\nsending last dtmf";
					pthread_mutex_lock(&mutex1);
					rtp_session->putData(timestamp, (unsigned char*)dtmf_payload, dtmf_payload_size);
					pthread_mutex_unlock(&mutex1);
					neat_tx_bytes+=dtmf_payload_size;
					gross_tx_bytes+=(dtmf_payload_size+40);
				}
				else if(cnt>0) {
					dtmf_payload->end = true;
					//cout << "\nsending a pause:  " << (unsigned char*)dtmf_payload;
					pthread_mutex_lock(&mutex1);
					rtp_session->putData(timestamp, (unsigned char*)dtmf_payload, dtmf_payload_size);
					pthread_mutex_unlock(&mutex1);
					neat_tx_bytes+=dtmf_payload_size;
					gross_tx_bytes+=(dtmf_payload_size+40);
				} 
				if(cnt==1) {
					timestamp +=  ((n_dtmf + n_pause) * nsamples);
					dtmf_send = false;
					cnt--;
				}
				cnt--;
			}
			else {
				pt = 0;
				rtp_session->setPayloadFormat(StaticPayloadFormat((ost::StaticPayloadType) pt));
				pthread_mutex_lock(&mutex1);
				rtp_session->putData(timestamp, payload, payload_size);
				pthread_mutex_unlock(&mutex1);
				neat_tx_bytes+=payload_size;
				gross_tx_bytes+=(payload_size+40);
				timestamp += nsamples;
				if (timestamp <= rtp_session->getCurrentTimestamp() - (JITTER_BUF_MS / ptime) * nsamples){
					timestamp += nsamples * (JITTER_BUF_MS / ptime);
				}
			}
		} 
		else {
			pt = 0;
			rtp_session->setPayloadFormat(StaticPayloadFormat((ost::StaticPayloadType) pt));
			pthread_mutex_lock(&mutex1);
			rtp_session->putData(timestamp, dummy, payload_size);
			pthread_mutex_unlock(&mutex1);
			neat_tx_bytes+=payload_size;
			gross_tx_bytes+=(payload_size+40);
			timestamp += nsamples;
			if (timestamp <= rtp_session->getCurrentTimestamp() - (JITTER_BUF_MS / ptime) * nsamples){
				timestamp += nsamples * (JITTER_BUF_MS / ptime);
			}
		}
		Thread::sleep(TimerPort::getTimer());
		TimerPort::incTimer(ptime);
	}
	is_running = false;
	delete[] dummy;
}


void audio_tx::close_socket() {
	terminate();
}

void audio_tx::stop_action() {
	active = false;
}


