/***************************************************************************
 *   Copyright (C) 2005 by ,,,   *
 *   giuseppe@Euclide   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef RTP_HANDLER_H
#define RTP_HANDLER_H


#include <assert.h>
#include <iostream>
#include <ctime>
#include <ccrtp/rtp.h>
#include <ccrtp/ext.h>
#include "ctrlcenter.h"
#include "DSbuffer.h"
#include "DSin.h"
#include "DSout.h"


#define RTP_PACK_SIZE 160  //payload bytes

using namespace ost;
using namespace std;

struct RtpDtmfPayload {
	unsigned char	event : 8;
	unsigned char	volume : 6;
	bool		reserved : 1;
	bool		end : 1;
	//duration in timestamp unit of a single packet
	unsigned short	duration : 16;
};

struct bit_wise_payload {
	unsigned char bit_payload : 1;
};

struct RtpDtmfEvent {
	DtmfEvent digit;
	//DTMF tone duration in timestamp unit
	unsigned int duration;
	// pause duration in timestamp unit
	unsigned int pause;
	unsigned int volume;
};

/**
* RTP Payload definitions 
sptPCMU 	ITU-T G.711. mu-law audio 8 Khz (RFC 1890).
sptG726_32 	ITU-T G.726. ADPCM audio (RFC 1890).
sptGSM 		GSM audio (RFC 1890).
sptG723 	ITU-T G.723. MP-MLQ ACELP audio (RFC 1890).
sptDVI4_8000 	Modified IMA ADPCM audio 8Khz (RFC 1890).
sptDVI4_16000 	Modified IMA ADPCM audio 16Khz (RFC 1890).
sptLPC 		LPC audio (RFC 1890).
sptPCMA 	ITU-T G.711 A-law audio 8 Khz (RFC 1890).
sptG722 	Audio (RFCs 1890, 3047).
sptL16_DUAL 	Linear uncompressed dual audio (RFC 1890).
sptL16_MONO 	Linear uncompressed mono audio (RFC 1890).
sptQCELP 	Audio at 8000 hz.
sptMPA 		MPEG Audio elem. stream (RFCs 1890, 2250).
sptG728 	ITU-T G.728. LD-CELP audio.
sptDVI4_11025 	DVI audio at 11025 hz (by Joseph Di Pol).
sptDVI4_22050 	DVI audio at 22050 hz (by Joseph Di Pol).
sptG729 	ITU-T G.729. CS-ACELP audio.
lastStaticAudioPayloadType 	
sptCELB 	Sun's propietary video (RFCs 1890, 2029).
sptJPEG 	JPEG (ISO 10918) video (RFCs 1890, 2435).
sptNV 	Ron Frederick's nv audio (RFC 1890).
sptH261 	ITU-T H.261 video (RFCs 1890, 2032).
sptMPV 	MPEG Video elem. stream (RFCs 1890, 2250).
sptMP2T 	MPEG 2 Transport stream (RFCs 1890, 2250).
sptH263 	ITU-T H.263 video (RFCs 2190, 2429).
*/

class RTP_rx: public Thread,
		     TimerPort
 {
private:
	// socket to transmit
	RTPSession* socket_rx;
	// host network address
	InetHostAddress local_ip;
	InetHostAddress remote_ip;

	// identifier of this sender
	uint32 ssrc;
	
	//Buffer to fill with data received from the network
	DSbuffer* Buffer;

	//
	bool RUN_FLAG;
	DSout out;
public:
	RTP_rx(DSbuffer* B , SessionPar* SP);
	~RTP_rx();
	
	void getrtp(DSbuffer* BUFF);
	
	void get_rtp_flow(DSbuffer* BUFF);

	void close_socket();

	void run(); 

	void newstart();
};


/**
 * @class RTP_Tx
 * Transmitter of salutes.
 */
class RTP_tx: public Thread, 
		     TimerPort
	     
{

private:
	// socket to transmit
	RTPSession* socket_tx;
	// loopback network address
	InetHostAddress remote_ip;
	InetHostAddress local_ip;

	// identifier of this sender
	uint32 ssrc;

	//buffer with data to be sent to the network
	DSbuffer* Buffer;
	bool RUN_FLAG;
	DSin in;
	

public:
	RTP_tx(DSbuffer* B, SessionPar* SP);

	~RTP_tx();

	void sendrtp (DSbuffer* BUFF);
	
	void send_rtp_flow(DSbuffer* BUFF, double RATE);
	
	void close_socket();

	void run();
	void newstart();
};


/**
 * @class RTP_Sym
 * Transmitter of salutes.
 */

class RTP_Sym : public SymmetricRTPSession
		      
{
public: 
	RTP_Sym( SessionPar* SP);
	~RTP_Sym();
	uint32 getLastTimestamp(const SyncSource *src=NULL) const;
		
private:
	// host network address
	InetHostAddress local_ip;
	InetHostAddress remote_ip;
	// identifier of this sender
	uint32 ssrc;
	
};

class audio_rx: public Thread, TimerPort
{
private:
	DSbuffer* buffer_rx;
	//Symmetric RTP session
	RTP_Sym *rtp_session;
	bool active;
	// Codec information
	unsigned short	ptime;	// in milliseconds

	// Buffer to store PCM samples for ptime ms
	unsigned char	*sample_buf;

	// RTP payload
	unsigned short payload_size;
	unsigned char *payload;
	unsigned short nsamples; // number of samples taken per packet

	// Inidicates if the recording thread is running
	bool is_running;

	// The thread exits when this indicator is set to true
	bool stop_running;

	// Timestamp for next RTP packet
	unsigned long timestamp;

	bool mute_flag;
	static double neat_rx_bytes;
	static double gross_rx_bytes;
	
public:
	// Create the audio receiver
	// _rtp_session	RTP socket to send the RTP stream
	// _ptime	length of the audio packets in ms
	// _ptime = 0 means use default ptime value for the codec
	audio_rx(RTP_Sym* _rtp_session, unsigned short _ptime = 0, DSbuffer* B = 0);

	~audio_rx();

	void run(void);
	void close_socket();
	void stop_action();
	void mute(bool M);
	void dtmf_sent(char P, int D);
	static double& get_neat_rx_bytes() {return neat_rx_bytes;}
	static double& get_gross_rx_bytes() {return gross_rx_bytes;}	

};

class audio_tx: public Thread, TimerPort
{

private:
	
	DSbuffer* buffer_tx;
	RTP_Sym* rtp_session;
	bool active;
	unsigned short	ptime;	// in milliseconds

	// Buffer to store PCM samples of a received RTP packet
	unsigned char	*sample_buf;

	// Jitter buffer (PCM).
	// jitter_buf_len indicates the number of bytes in the jitter buffer.
	// At the start of playing the samples are stored in the jitter buffer.
	// Once the buffer is full, all samples are copied to the memory of
	// the soundcard. From that point the soundcard itself is the jitter
	// buffer.
	unsigned char	*jitter_buf;
	unsigned short	jitter_buf_len;
	bool		load_jitter_buf;
	
	// RTP payload
	unsigned short payload_size;
	unsigned char *payload;
	unsigned short nsamples; 

	// RTP DTMF Event payload
	unsigned short dtmf_payload_size;
	RtpDtmfPayload *dtmf_payload;

	// Inidicates if the playing thread is running
	bool is_running;

	// The thread exits when this indicator is set to true
	bool stop_running;

	// Timestamp for next RTP packet
	unsigned long timestamp;
	
	bool dtmf_send;
	int n_dtmf;
	int n_pause;
	RtpDtmfPayload event_payload;
	
	bool mute_flag;
	static double neat_tx_bytes;
	static double gross_tx_bytes;

public:
	// Create the audio transmitter
	
	// _rtp_session	RTP socket tp send the RTP stream
	// _ptime	length of the audio packets in ms
	// _ptime = 0 means use default ptime value for the codec
	audio_tx(RTP_Sym* _rtp_session, unsigned short _ptime = 0, DSbuffer* B = 0);

	~audio_tx();

	void run(void);
	void close_socket();
	void stop_action();
	bool send_dtmf(RtpDtmfEvent* event);
	bool send_dtmf_in_band(char P, int D);
	void mute(bool M);
	static double& get_neat_tx_bytes() {return neat_tx_bytes;}
	static double& get_gross_tx_bytes() {return gross_tx_bytes;}
};


#endif
