/***************************************************************************
 *   Copyright (C) 2005 by ,,,   *
 *   giuseppe@Euclide   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef USERAGENTSIP_H
#define USERAGENTSIP_H

/**
Main Class to handle SIP sessions 
@author ,,,
*/
/*#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
*/
#include "resiprocate/NameAddr.hxx"
#include "resiprocate/SdpContents.hxx"
#include "resiprocate/StackThread.hxx"
#include "resiprocate/dum/ClientAuthManager.hxx"
#include "resiprocate/dum/AppDialogSet.hxx"
#include "resiprocate/dum/ClientInviteSession.hxx"
#include "resiprocate/dum/ClientRegistration.hxx"
#include "resiprocate/dum/DialogSetHandler.hxx"
#include "resiprocate/dum/DialogUsageManager.hxx"
#include "resiprocate/dum/DumThread.hxx"
#include "resiprocate/dum/InviteSessionHandler.hxx"
#include "resiprocate/dum/MasterProfile.hxx"
#include "resiprocate/dum/OutOfDialogHandler.hxx"
#include "resiprocate/dum/RegistrationHandler.hxx"
#include "resiprocate/dum/ServerInviteSession.hxx"
#include "resiprocate/os/Logger.hxx"
#include "resiprocate/os/Random.hxx"


namespace resip {

//using namespace resip;

#define RESIPROCATE_SUBSYSTEM Subsystem::SIP


class UserAgentSip : public ClientRegistrationHandler,
		     public OutOfDialogHandler,
                     public InviteSessionHandler,
		     public DialogSetHandler

{
    
public:
	UserAgentSip();
	UserAgentSip( NameAddr Aor,  Data password, const char* ptime, Data sipport);
	virtual ~UserAgentSip();

	void update( NameAddr Aor,  Data password, const char* ptime, Data sipport);
	void registrationSetup(NameAddr Aor, const Data password, const Data regDuration);
	void initializeCodecList();
	SdpContents setSdp(const char* mPtime);
	SdpContents setSdp(const char* mPtime, int Port);
	void process();
	void outgoingCall(Uri &callee, const char* ptime);
	void endingCall(int);
	void deleteRegistration();
	void shutdown();
	void setPtime(const char* ptime);
	void setPort(Data port);
	int getPort() {return mPort;}
	int getRtpPort() {return mRtpPort;}
	Data get_IP() {return mStack.getHostAddress();}

public:
	int& get_MM_SessionID() {return MM_SessionID;}
       // Invite Session Handler /////////////////////////////////////////////////////
      virtual void onNewSession(ClientInviteSessionHandle h, InviteSession::OfferAnswerType oat, const SipMessage& msg);
      virtual void onNewSession(ServerInviteSessionHandle h, InviteSession::OfferAnswerType oat, const SipMessage& msg);
      virtual void onFailure(ClientInviteSessionHandle h, const SipMessage& msg);
      virtual void onEarlyMedia(ClientInviteSessionHandle, const SipMessage&, const SdpContents&);
      virtual void onProvisional(ClientInviteSessionHandle, const SipMessage& msg);
      virtual void onConnected(ClientInviteSessionHandle h, const SipMessage& msg);
      virtual void onConnected(InviteSessionHandle, const SipMessage& msg);
      virtual void onStaleCallTimeout(ClientInviteSessionHandle);
      virtual void onTerminated(InviteSessionHandle h, InviteSessionHandler::TerminatedReason reason, const SipMessage* msg);
      virtual void onRedirected(ClientInviteSessionHandle, const SipMessage& msg);
      virtual void onAnswer(InviteSessionHandle, const SipMessage& msg, const SdpContents&);
      virtual void onOffer(InviteSessionHandle handle, const SipMessage& msg, const SdpContents& offer);
      virtual void onOfferRequired(InviteSessionHandle, const SipMessage& msg);
      virtual void onOfferRejected(InviteSessionHandle, const SipMessage& msg);
      virtual void onDialogModified(InviteSessionHandle, InviteSession::OfferAnswerType oat, const SipMessage& msg);
      virtual void onInfo(InviteSessionHandle, const SipMessage& msg);
      virtual void onInfoSuccess(InviteSessionHandle, const SipMessage& msg);
      virtual void onInfoFailure(InviteSessionHandle, const SipMessage& msg);
      virtual void onRefer(InviteSessionHandle, ServerSubscriptionHandle, const SipMessage& msg);
      virtual void onReferAccepted(InviteSessionHandle, ClientSubscriptionHandle, const SipMessage& msg);
      virtual void onReferRejected(InviteSessionHandle, const SipMessage& msg);

      // Registration Handler ////////////////////////////////////////////////////////
      virtual void onSuccess(ClientRegistrationHandle h, const SipMessage& response);
      virtual void onFailure(ClientRegistrationHandle h, const SipMessage& response);
      virtual void onRemoved(ClientRegistrationHandle h);
      virtual int onRequestRetry(ClientRegistrationHandle h, int retryMinimum, const SipMessage& msg);

      // OutOfDialogHandler //////////////////////////////////////////////////////////
      virtual void onSuccess(ClientOutOfDialogReqHandle, const SipMessage& response);
      virtual void onFailure(ClientOutOfDialogReqHandle, const SipMessage& response);
      virtual void onReceivedRequest(ServerOutOfDialogReqHandle, const SipMessage& request);
      virtual void onForkDestroyed(ClientInviteSessionHandle);
      
      //called when a 100 is received 
      virtual void onTrying(AppDialogSetHandle, const SipMessage& msg);
      
      //called when a provisional response that does not create a dialog arrives
      //usually, this is a 180 w/out a contact
      virtual void onNonDialogCreatingProvisional(AppDialogSetHandle, const SipMessage& msg);
	
	void addCodec(SdpContents::Session::Medium* MD, int CODE);
	void addMedium(SdpContents* SDP, const char* TYPE, std::list<int>* CODE_LIST, int PORT);
	bool parseSdp(SdpContents* sdp, void* SP);

protected:
   

private:
	SipStack mStack;
	DialogUsageManager mDum;
	StackThread mStackThread;
	DumThread mDumThread;
	MasterProfile mProfile;
 
      SipMessage* mInvite;
      SipMessage mIncomingInvite;

      ClientRegistrationHandle mCr;
      Data mDialogid;
      Data mLogType;
      Data mLogLevel;
      int mPort;
      int mRegisterDuration;
      Uri mAor;
      Uri mTo;
      NameAddr mContact;
      Data mPassword;
      int mRtpPort;
      const char* mPtime;
      int MM_SessionID;
	SdpContents::Session::Codec::CodecMap* SupportedCodecs;
  };
}

    void* WhileFunction(void *ptr);
    void* RemoveFunction(void* ptr);




#endif
