/*********************************************************************************
 This file includes the definition of following classes:
 1) CCThread -- the control tread defined above the POSIX thread 
 2) ControlStatus -- the class that manages the status of the software

 Also included:
 3) function "CtrlFunction" -- code to be executed by each CCThread
 4) global variables initialization
*********************************************************************************/



#ifndef _CTRLCENTER_H 
#define _CTRLCENTER_H

//#include <stdio.h>
//#include <stdlib.h>
//#include <string.h>
//#include <unistd.h>
#include <iostream>
#include <qtextedit.h>
#include <sys/time.h>
#include <pthread.h>
#include <queue_t.h>
#include "useragentwin.h"

#include <cc++/socket.h>
#include "DSin.h"
#include "DSout.h"
#include "DSbuffer.h"
#include <qdatetime.h>

#define PACKAGE_SOUND_DIR "/usr/share/sounds/RingTones/"

using namespace ost;
class audio_rx;
class audio_tx;
class RTP_Sym;

enum sip_status {
	idle,
	active,
	invite_sent,
	invite_received,
	cancelled,
	rejected
} ;

enum DtmfEvent {
	DTMFdigit0,
	DTMFdigit1,
	DTMFdigit2,
	DTMFdigit3,
	DTMFdigit4,
	DTMFdigit5,
	DTMFdigit6,
	DTMFdigit7,
	DTMFdigit8,
	DTMFdigit9,
	DTMFdigitstar,
	DTMFdigitsharp,
	DTMFdigitA,
	DTMFdigitB,
	DTMFdigitC,
	DTMFdigitD
} ;


typedef struct {
int SSID;
char CID[80]; 
} Record;


typedef struct {
//session identifier
	int id;
	int pos_id;
	sip_status session_status;
//Media parameters:
	int media_type;
	bool mute;
	DSin* Remote;
	DSout* Locale;
	DSbuffer* buffer_a2n;
	DSbuffer* buffer_n2a;
//RTP parameters:
	InetHostAddress dest_IP;
	InetHostAddress source_IP;
	audio_rx* ARX;
	audio_tx* ATX;
	RTP_Sym* RTP_SYM;
	int source_port;
	int dest_port;
//CODEC specifications:
	char codec_name[40];
	unsigned long b; //bandwidth in kbps
	unsigned short ptime; //packet time in ms
	double bpf; //bytes per frame
	double fpp; //frames per packet
//UA handlers
	ClientInviteSessionHandle* mCis;
	ServerInviteSessionHandle* mSis;
//counter
	QTime start_time;
} SessionPar;

class ControlStatus {
public:
	int session_counter;
	int active_audio;
	int active_video;
	int active_chat;
	
	bool busy;
	bool registered;
	bool* ringback;
	bool* ringtone;
	bool* busytone;
	bool* failtone;
	int mode; //operational mode: 1-Normal, 2-Offer/Answere, 3-QoS_Negotiation
	queue<SessionPar>* SessionList;
	queue<Record>* TableList;
	UserAgentWin *Win;
	Record R;
	
public:
	ControlStatus();
	~ControlStatus(){}
	void append(const string s) {Win->teDisplay->append(s);}
	void initialize(UserAgentWin *W);
	void set_mode(int M);
	int get_mode();
	UserAgentWin* getWin() {return Win;}
	queue<SessionPar>* get_list() {return SessionList;}
	queue<Record>* get_table() {return TableList;}
	bool get_SSID(char* cid, int* ssid);
	bool get_CID(int ssid, char** cid, int* len);

	bool* get_ringback(int p);
	bool* get_ringtone(int p);
	bool* get_busytone(int p);
	bool* get_failtone(int p);

	bool new_outgoing_req(SessionPar* SP); //checks whether to proceed with an outgoing INVITE or not
	bool new_incoming_req(SessionPar* SP); //checks whether to proceed with an ingoing INVITE or not
	int add_session(SessionPar* SP); //session to be added: store it in the SessionList and return the SessionID
	void start_session(SessionPar* SP);
	bool remove_session(int SessionID);
	SessionPar* get_session(int ID);
	SessionPar* get_session(char* CID);
	void start_ringback(int);
	void start_ringtone(int);
 	void start_busytone(int);
	void start_failtone(int);
	void stop_ringing(int);
 	void stop_busytone(int);
	void stop_failtone(int);

	bool start_media(SessionPar* SP);
	bool start_media_session(SessionPar* SP); // fills "buffer_in" and "buffer_out" fields
	bool start_media_session(int SessionID);
	bool stop_media_session(SessionPar* SP);
	bool stop_media_session(int SessionID);

	bool start_RTP_session(SessionPar* SP); //fills "buffer_rx" and "buffer_tx" fields 
	bool start_RTP_session(int SessionID);
	bool stop_RTP_session(SessionPar* SP);
	bool stop_RTP_session(int SessionID);
	double get_neat_rx_bytes();
	double get_neat_tx_bytes();
	double get_gross_rx_bytes();
	double get_gross_tx_bytes();
	void set_rx_bytes(double, double);
	void set_tx_bytes(double, double);
};

class CCThread {
protected:
	ControlStatus* Status;
	void* (*Function) (void*);
	void* Parameters;
	pthread_t THR;
public:
	CCThread(ControlStatus* Stat, void * (*Func)(void *), void* Par);
	~CCThread() {}
	int run();
	void join();
	void exit(int*);
};

void *CtrlFunction( void *ptr );
void *SoundFunction ( void *ptr );
void* RecFunction(void* ptr);
void* PlayFunction(void* ptr);
extern pthread_mutex_t mutex1;
extern pthread_mutex_t mutex2;
extern pthread_mutex_t mutex3;
extern int counter;
extern ControlStatus SoftState;


#endif
