/***************************************************************************
 *   Copyright (C) 2005 by ,,,   *
 *   giuseppe@Euclide   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "useragentsip.h"
#include <iostream>
#include <qlabel.h>
#include <qdatetime.h>
#include <kmainwindow.h>
#include <klocale.h>
#include "ctrlcenter.h"
#include "rtp_handler.h"
#include <assert.h>
#include <sys/socket.h>

using namespace std;
using namespace resip;
using namespace ost;

typedef struct {
	SessionPar* SP;
	const char** mPtime;
	ServerInviteSessionHandle* h; 
	UserAgentSip* UAS;
} session_data;

session_data SD;

UserAgentSip::UserAgentSip() :
	mStack(),
	mDum(mStack),
	mStackThread(mStack),
	mDumThread(mDum),
	mProfile()
{
printf("\nVDB--> UserAgentSip::UserAgentSip()\n");
	mInvite = NULL;
}


UserAgentSip::UserAgentSip( NameAddr Aor, Data password , const char* ptime, Data sipport ) :
	mStack(),
	mDum(mStack),
	mStackThread(mStack),
	mDumThread(mDum),
	mProfile()
	
{
printf("\nVDB--> UserAgentSip::UserAgentSip( NameAddr Aor, Data password , const char* ptime, Data sipport )\n");
	//SIP contact URI is the address (or URI) to a phone.
	//This usually consists of a device name, IP address and port number. 
	//This is a temporary URI that your SIP location server keeps in memory as long as it is valid. 
	mInvite = NULL;
	Uri contact;
	mPort = sipport.convertInt();
	cout << "\nVDB --> mPort = " << mPort << endl;

	mRtpPort = 50000;
	cout << "\nVDB --> Using RTP port " << mRtpPort; 
	//SIP AOR (Adress of record) URI is the permanent address. 
	//This is the address people can use to call you, 
	//wherever you are, regardless of IP adress or phone.
	//Your phone registers with the location server to tell it how to map one permanent (AOR) address 
	//to one or several SIP contacts. 
	mAor = Aor.uri();
	mPassword = password;
	mPtime = ptime;

	contact.user() = mAor.user();
        contact.host() = mStack.getHostAddress();

	if(mPort == 0) mPort = 5060;
	int max_Port = mPort+200;
	int sock;
	struct sockaddr_in addr;

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	if((sock = socket(AF_INET,SOCK_DGRAM,0))<0) {
			cout << "\nVDB --> Cannot open a new socket" << endl;
			exit(-1);
	}
	while(mPort < max_Port) {
		addr.sin_port = htons(mPort);	
		if(bind(sock,(struct sockaddr*)&addr,sizeof(struct sockaddr))<0) {
			cout << "\nVDB --> Port " << mPort << " not available" << endl;
			mPort++;
		}
		else {
			cout << "\nVDB --> Using port " << mPort << endl; 
			break;
		}
	}
	if(mPort >= max_Port) {	
		cout << "\n\nVDB --> No port available in the range " << mPort-200 << " - " << mPort << endl;
		exit (-1);
	}
	
	close(sock);
	
	mDum.addTransport(UDP, mPort, V4);

       	contact.port() = mPort;
	mProfile.setOverrideHostAndPort(contact);
	mProfile.setUserAgent("TTI_SUA v.4.2");

	mAor.port() = mPort;
	cout <<"\nVDB--> mAor = " << mAor << endl;
	mProfile.setDefaultFrom(NameAddr(mAor));
   	mProfile.setDigestCredential(mAor.host(), mAor.user(), mPassword);

	mDum.setMasterProfile(&mProfile);
	mDum.setDialogSetHandler(this);
	mDum.setClientRegistrationHandler(this);
	mDum.setInviteSessionHandler(this);
	mStackThread.run();
	mDumThread.run();
}

UserAgentSip::~UserAgentSip()
{
	printf("\nVDB--> UserAgentSip::~UserAgentSip()\n");
	deleteRegistration();
	mStackThread.shutdown();
  	mStackThread.join();
	mDumThread.shutdown();
	mDumThread.join();

}


void UserAgentSip::update( NameAddr Aor, Data password , const char* ptime, Data sipport )
{
printf("\nVDB--> UserAgentSip::update( NameAddr Aor, Data password , const char* ptime, Data sipport )\n");
	Uri contact;
	mAor = Aor.uri();
	mPassword = password;
	mPtime = ptime;
	contact.user() = mAor.user();
        contact.host() = mStack.getHostAddress();
	contact.port() = mPort;
	if(mPort != sipport.convertInt()) {
		mPort = sipport.convertInt();
		if(mPort == 0) mPort = 5060;
		int max_Port = mPort+200;
		int sock;
		struct sockaddr_in addr;
		addr.sin_family = AF_INET;
		addr.sin_addr.s_addr = INADDR_ANY;
		if((sock = socket(AF_INET,SOCK_DGRAM,0))<0) {
			cout << "\nVDB --> Cannot open a new socket" << endl;
			exit(-1);
		}
		while(mPort < max_Port) {
			addr.sin_port = htons(mPort);	
			if(bind(sock,(struct sockaddr*)&addr,sizeof(struct sockaddr))<0) {
				cout << "\nVDB --> Port " << mPort << " not available" << endl;
				mPort++;
			}
			else {
				cout << "\nVDB --> Using port " << mPort << endl; 
				break;
			}
		}
		if(mPort >= max_Port) {	
			cout << "\n\nVDB --> No port available in the range " << mPort-200 << " - " << mPort << endl;
			exit (-1);
		}
		close(sock);
		mDum.addTransport(UDP, mPort, V4);
	       	contact.port() = mPort;
	}
	mProfile.setOverrideHostAndPort(contact);
	mAor.port() = mPort;
	cout <<"\nVDB--> mAor = " << mAor << endl;
	mProfile.setDefaultFrom(NameAddr(mAor));
   	mProfile.setDigestCredential(mAor.host(), mAor.user(), mPassword);

}

void 
UserAgentSip::registrationSetup( NameAddr Aor, const Data password , const Data regDuration )
{
printf("\nVDB--> UserAgentSip::registartionSetup( NameAddr Aor, const Data password , const Data regDuration )\n");
	mAor = Aor.uri();
//	mContact = Aor;

	mPassword = password;
	//The RegisterDuration parameter indicates how long the URI is valid.
	//The "expires" header field in the Register message is set to this value.
	mRegisterDuration = regDuration.convertInt();
	(mDum.getMasterProfile())->setDefaultRegistrationTime(mRegisterDuration);
	
	auto_ptr<ClientAuthManager> cam(new ClientAuthManager);
	mDum.setClientAuthManager(cam);
	mProfile.setDigestCredential(mAor.host(), mAor.user(), mPassword);
	if(!mRegisterDuration){
		//mCr->removeBinding(mContact);
		mCr->end();
	}
	else{	// first registration
  		SipMessage reg = mDum.makeRegistration( NameAddr(mAor), mProfile );
		mDum.send(reg);
	}
}

SdpContents
UserAgentSip::setSdp(const char* mPtime)
{
	printf("\nVDB--> UserAgentSip::setSdp(const char* mPtime)\n");
	SdpContents sdp;
	SdpContents::Session::Medium audio("audio", mRtpPort, 0, "RTP/AVP");
	//<bandwidth-value> is in kilobits per second
	//<modifier> Two modifiers are initially defined: 1)CT Conference Total and AS Application-Specific Maximum
	SdpContents::Session::Bandwidth bandwidth("AS",64);
	// Setting codec G.711
	audio.addCodec(SdpContents::Session::Codec("PCMU", 0, 8000));
	audio.addAttribute("ptime", mPtime);
	audio.addBandwidth(bandwidth);
	// Inizializzazione dell'SDP
	sdp.init();
	sdp.session().version() = 0;
	sdp.session().origin().user() = "TTI";
	sdp.session().origin().setAddress( mStack.getHostAddress() , SdpContents::IP4 );
	sdp.session().connection().setAddress( mStack.getHostAddress() , SdpContents::IP4 );
	sdp.session().name() = "Session SDP";
	sdp.session().addMedium(audio);
	return sdp;
}

SdpContents
UserAgentSip::setSdp(const char* mPtime, int Port)
{
	printf("\nVDB--> UserAgentSip::setSdp(const char* mPtime, unsigned Port)\n");
	SdpContents sdp;
	int temp_port = mRtpPort;
	mRtpPort=Port;
	sdp=setSdp(mPtime);
	mRtpPort=temp_port;
	return sdp;
}

void 
UserAgentSip::outgoingCall(Uri &callee, const char* ptime)
{
	printf("\nVDB--> UserAgentSip::outgoingCall(Uri &callee, const char* ptime)\n");
	Uri cnt;
        NameAddr to, from;
	SoftState.Win->LoadConfig();
	SessionPar* SP = new SessionPar;
	SP->session_status = idle;
	SP->start_time=QTime::currentTime(); 
	SP->pos_id = SoftState.Win->get_console();
	SP->mSis = NULL;
	SP->mCis = NULL;	
	SP->media_type=0;
	if(SoftState.get_list()->get_npack()==0)
		SP->mute = false;
	else 
		SP->mute = true;
	SP->source_IP= (mStack.getHostAddress()).c_str();
	SP->ARX=NULL;
	SP->ATX=NULL;
	SP->RTP_SYM = NULL;
	SP->source_port = mRtpPort+2*((Random::getRandom())%5000);
	SP->dest_port = -1;
//	SP->dest_IP=(((sdp->session()).connection()).getAddress()).c_str();	
	strcpy(SP->codec_name,"PCMU");
//	SP->b=(((sdp->session()).bandwidths()).front()).kbPerSecond();//bandwidth in kbps; //bandwidth in bps
	SP->ptime= atoi(mPtime); //packet time in ms
	SP->bpf=1; //bytes per frame
	SP->fpp=1; //frames per packet
// la seguente chiamata a funzione non fa partire i media e RTP perch lo stato  idle
	Record* R = new Record();
	R->SSID = SoftState.add_session(SP);	
	
	if(!SoftState.registered){
	        cnt.user() = mAor.user();
		cnt.host() = mStack.getHostAddress();
//        	contact.uri() = cnt;
		from.uri() = cnt;
		(mDum.getMasterProfile())->setDefaultFrom(NameAddr(from));	
	}

	mTo = callee;
	to.uri() = mTo;
	mPtime = ptime;
	to.uri().param(p_transport) = "udp";
	from.uri().param(p_transport) = "udp";
	SdpContents sdp_temp = setSdp(mPtime, SP->source_port);
	SipMessage &mTempInvite = mDum.makeInviteSession(to, mProfile, &sdp_temp);
	mInvite = &mTempInvite;
	H_CallID *cid = new H_CallID();
	H_CallID& rc = *cid;
	char message[80];
	sprintf(message,"%s",(char*)mTempInvite.header(rc).value().data());
	strcpy(R->CID, message); 
	SoftState.get_table()->push(R);

	mDum.send(mTempInvite);
	SP->session_status = invite_sent;
	}

void
UserAgentSip::endingCall(int LID)
{
printf("\nVDB--> UserAgentSip::endingCall()\n");
	
	SessionPar* SP = SoftState.get_session(LID);

	char parola[20];
	sprintf(parola,"status = %d",SP->session_status);
	printf("\nVDB--> %s\n",parola);

	if(SP->session_status == invite_received)
		SP->session_status = rejected;
	else if(SP->session_status == active) {
		if(SP->mSis != NULL) {
			(*(SP->mSis))->end();
			//SP->mSis=NULL;
		} else if(SP->mCis != NULL) {
			(*(SP->mCis))->end();
			//SP->mCis=NULL;
		} else {
			printf("Error");
			exit(-1);
		}
	} else if(SP->session_status == invite_sent) {
		//send CANCEL
		SP->session_status = cancelled;
		if(SP->mCis != NULL) {
			AppDialogSetHandle s = (*(SP->mCis))->getAppDialogSet();
 			if(s.get()!=NULL) 
				s.get()->end();
			SP->mCis=NULL;
		}
		//SoftState.remove_session(SP->id);
	}
	
	SoftState.stop_ringing(SP->pos_id);
	SoftState.stop_busytone(SP->pos_id);
	SoftState.stop_failtone(SP->pos_id);
}

/// Not used
void
UserAgentSip::deleteRegistration(){
printf("\nVDB--> UserAgentSip::deleteRegistration()\n");
	if(!mRegisterDuration){
		InfoLog( << "Deregistrating user:" << mContact );	
		mCr->end();
	}
}

/// Not used
void 
UserAgentSip::process()
{
	printf("\nVDB--> UserAgentSip::process()\n");
	FdSet fdset;
	mStack.buildFdSet(fdset);
	int err = fdset.selectMilliSeconds(20);
        assert ( err != -1 );
	mStack.process(fdset);
	while (mDum.process());
}

/// Not used
void 
UserAgentSip::shutdown()
{
	printf("\nVDB--> UserAgentSip::shutdown()\n");
} 

void 
UserAgentSip::setPort(Data sipport)
{
	printf("\nVDB--> UserAgentSip::setPort(Data sipport)\n");
	mPort=sipport.convertInt();
} 

void 
UserAgentSip::setPtime(const char* ptime)
{
	printf("\nVDB--> UserAgentSip::setPtime(const char* ptime)\n");
	mPtime=ptime;
} 

///called when a 100 is received 
void 
UserAgentSip::onTrying(AppDialogSetHandle, const SipMessage& msg)
{
	printf("\nVDB--> UserAgentSip::onTrying(AppDialogSetHandle, const SipMessage& msg)\n");
	int code = 0;
	sscanf((char*)msg.brief().data(),"%*s %d %*s",&code);
	SoftState.append("  " + (string)Data(code).c_str() + "  Trying");
}

      
///called when a provisional response that does not create a dialog arrives
///usually, this is a 180 w/out a contact
void 
UserAgentSip::onNonDialogCreatingProvisional(AppDialogSetHandle, const SipMessage&)
{
	printf("\nVDB--> UserAgentSip::onNonDialogCreatingProvisional(AppDialogSetHandle, const SipMessage& msg)\n");
}


void
UserAgentSip::onNewSession(ClientInviteSessionHandle h, InviteSession::OfferAnswerType oat, const SipMessage& msg)
{
 	printf("\nVDB--> UserAgentSip::onNewSession(ClientInviteSessionHandle h, InviteSession::OfferAnswerType oat, const SipMessage& msg) %p %p\n", &oat, &msg);
	
	InfoLog(<< h->myAddr().uri().user() << " 180 from  " << h->peerAddr().uri().user());

	int code = 0;
	sscanf((char*)msg.brief().data(),"%*s %d %*s",&code);
	SoftState.append("  " + (string)Data(code).c_str() + "  Ringing");
		
	SessionPar* SP = SoftState.get_session((char*)h->getDialogId().data());
	assert(SP != NULL);
	printf("\nVDB--> UserAgentSip::onNewSession status=%d \n",SP->session_status );
 
	if(SP->session_status == invite_sent) {
		/// start ringing
		SoftState.start_ringback(SP->pos_id);
		SP->mCis = new ClientInviteSessionHandle();
		*(SP->mCis) = h; //->getSessionHandle());
		if(SoftState.Win->get_console()!=SoftState.Win->get_line()) {
			SP->mute = true;
		}
		else {
			SoftState.Win->set_mute(SP->mute);
		}
		
		
	}
	else {
		SoftState.append("\tErron, no such requested call in progress");
		if(SP!=NULL)
			SoftState.remove_session(SP->id);
	}
}

void
UserAgentSip::onNewSession(ServerInviteSessionHandle h, InviteSession::OfferAnswerType, const SipMessage& msg)
{
//Empty INVITE not supported, for now

	printf("\nVDB--> UserAgentSip::onNewSession(ServerInviteSessionHandle h, InviteSession::OfferAnswerType oat, const SipMessage& msg)\n");

	InfoLog(<< h->myAddr().uri().user() << " INVITE from  " << h->peerAddr().uri().user());
	SoftState.Win->LoadConfig();
	mIncomingInvite = msg;
	SdpContents* sdpInvite = dynamic_cast<SdpContents*>(msg.getContents());
	
	char message[80];
	strcpy(message,"System Time: ");
	strcat(message,(QTime::currentTime()).toString("hh:mm:ss"));
	SoftState.append(message);
	strcpy(message,"  INVITE from: ");
	strcat(message,(char*)h->peerAddr().uri().user().data());
	strcat(message,"@");
	strcat(message,(char*)h->peerAddr().uri().host().data());	

	SessionPar* SP;
   	if(SoftState.busy) {
		cout << "\nVDB busy!" << endl;
		h->reject(486);
		cout << "\trejected!" << endl;
		strcat(message, " --> REFUSED");
		SoftState.append(message);
	}	
	else {
		SoftState.append(message);
		SoftState.new_incoming_req(0);
		int k=0;
		bool flag = false;
		while(!flag && k<SoftState.get_list()->get_len()) {
			k++;
			int j;
			for(j=0 ; j<SoftState.get_list()->get_npack(); j++)
				if((*SoftState.get_list())[j]->pos_id==k) {
					j=-1;
					break;
				}
			if(j!=-1)
				flag = true;
		}
		if(flag) {
			assert(k>0 && k<=SoftState.get_list()->get_len());
		}
		else {
			cout << "\nError: not busy but no line available!" << endl;
			exit(-1);
		}

		cout << "\nVDB --> Line=" << k << endl;

		SP = new SessionPar;
		SP->start_time=QTime::currentTime();
		SP->session_status = invite_received;
		SP->pos_id = k;
		SP->media_type=0;
		if(SoftState.get_list()->get_npack()==0)
			SP->mute = false;
		else 
			SP->mute = true;
		SP->source_IP =(mStack.getHostAddress()).c_str();
		SP->mCis = NULL;
		SP->mSis = NULL;
		SP->ATX = NULL;
		SP->ARX = NULL;
		SP->RTP_SYM = NULL;
		SP->source_port = mRtpPort+2*((Random::getRandom())%5000);;
		SP->dest_port = (((sdpInvite->session()).media()).front()).port();
		SP->dest_IP=(((sdpInvite->session()).connection()).getAddress()).c_str();	
		strcpy(SP->codec_name,"PCMU");
		SP->b=(((sdpInvite->session()).bandwidths()).front()).kbPerSecond();//bandwidth in kbps
		SP->ptime= atoi(mPtime); //packet time in ms
		SP->bpf=1; //bytes per frame
		SP->fpp=1; //frames per packet
		Record *R = new Record();
		R->SSID = SoftState.add_session(SP);

		h->provisional(180);
		SoftState.start_ringtone(SP->pos_id);
		mDialogid = (h->getDialogId());
		strcpy(R->CID, (char*) mDialogid.data());
		SoftState.TableList->push(R);

		SP->mSis = new ServerInviteSessionHandle();
		*(SP->mSis) = h; 
		//*(SP->mSis) = h->getSessionHandle();

		SD.SP = SP;
		SD.mPtime = &mPtime;
		SD.h = &h;
		cout << "\n VDB THIS = " << this << endl; 
 		SD.UAS = this;

		CCThread while_thread(&SoftState, WhileFunction, &SD);
		while_thread.run();
		
		SoftState.Win->set_active(SP->pos_id);
   }
}

void
UserAgentSip::onFailure(ClientInviteSessionHandle h, const SipMessage& msg)
{
	printf("\nVDB--> UserAgentSip::onFailure(ClientInviteSessionHandle h, const SipMessage& msg)\n");
   	InfoLog(<< h->myAddr().uri().user() 
           << " outgoing call failed " 
           << h->peerAddr().uri().user() 
           << " status=" << msg.header(h_StatusLine).statusCode());
   	InfoLog(<< "onFailure ");
	SessionPar* SP = SoftState.get_session((char*)h->getDialogId().data());	
	int code = 0;
	sscanf((char*)msg.brief().data(),"%*s %d %*s",&code);
	switch (code) {
		case 402:
			SoftState.append("  402 Payment required");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 403:
			SoftState.append("  403 Forbidden");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 404:
			SoftState.append("  404 Not Found");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 405:
			SoftState.append("  405 Method not allowed");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 406:
			SoftState.append("  406 Not acceptable");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 407:
			SoftState.append("  407 Proxy authentication required");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 408:
			SoftState.append("  408 Request timeout");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 409:
			SoftState.append("  409 Conflict");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 410:
			SoftState.append("  410 Gone");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 411:
			SoftState.append("  411 Length required");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 413:
			SoftState.append("  413 Request entity too lon");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 414:
			SoftState.append("  414 Request URI (URL) too long");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 415:
			SoftState.append("  415 Unsupported media type");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 420:
			SoftState.append("  420 Bad extension");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 480:
			SoftState.append("  480 Temporary unavailable");	
			SoftState.start_busytone(SP->pos_id);
			break;
		case 481:
			SoftState.append("  481 Call leg does not exist");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 482:
			SoftState.append("  482 Loop detected");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 483:
			SoftState.append("  483 Too many hops");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 484:
			SoftState.append("  484 Address incomplete");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 485:
			SoftState.append("  485 Address ambiguous");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 486:
			SoftState.append("  486 Busy here");	
			SoftState.start_busytone(SP->pos_id);
			break;
		case 487:
			SoftState.append("  487 Request cancelled");	
			SoftState.start_busytone(SP->pos_id);
			break;
		case 488:
			SoftState.append("  488 Not acceptable here");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 500:
			SoftState.append("  500 Internal server error");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 501:
			SoftState.append("  501 Not implemented");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 502:
			SoftState.append("  502 Bad gateway");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 503:
			SoftState.append("  503 Service unavailable");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 504:
			SoftState.append("  504 Gateway timeout");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 505:
			SoftState.append("  505 Version not implemented");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 580:
			SoftState.append("  580 Precondition Failed");
			SoftState.start_failtone(SP->pos_id);
			break;
		case 600:
			SoftState.append("  600 Busy everywhere");	
			SoftState.start_busytone(SP->pos_id);
			break;
		case 603:
			SoftState.append("  603 Decline");	
			SoftState.start_busytone(SP->pos_id);
			break;
		case 604:
			SoftState.append("  604 Does not exist anywhere");	
			SoftState.start_failtone(SP->pos_id);
			break;
		case 606:
			SoftState.append("  606 Not acceptable");
			SoftState.start_failtone(SP->pos_id);
			break;
		default:
			SoftState.append("  Unknown Error Code");	
			SoftState.start_failtone(SP->pos_id);
		}
}
      
void
UserAgentSip::onEarlyMedia(ClientInviteSessionHandle, const SipMessage&, const SdpContents&)
{
	printf("\nVDB--> UserAgentSip::onEarlyMedia(ClientInviteSessionHandle, const SipMessage&, const SdpContents&)\n");
	InfoLog (<< "onEarlyMedia ");
}

void
UserAgentSip::onProvisional(ClientInviteSessionHandle, const SipMessage&)
{
	printf("\nVDB--> UserAgentSip::onProvisional(ClientInviteSessionHandle, const SipMessage& msg)\n");
	InfoLog(<< "onProvisional ");
}

void
UserAgentSip::onConnected(ClientInviteSessionHandle h, const SipMessage& msg)
{
	printf("\nVDB--> UserAgentSip::onConnected(ClientInviteSessionHandle h, const SipMessage& msg)\n");
	
	SdpContents* sdp = dynamic_cast<SdpContents*>(msg.getContents());
   	InfoLog(<< h->myAddr().uri().user() << " in INVITE session with " << h->peerAddr().uri().user());

	int code;
	char message[32];
	sscanf((char*)msg.brief().data(),"%*s %d %*s",&code);
	sprintf(message,"  %d OK - On Line",code);

///fai partire connessione RTP finch non termina la chiamata

	SessionPar* SP = SoftState.get_session((char*)h->getDialogId().data());
	assert(SP!=NULL);
	SoftState.stop_ringing(SP->pos_id);
	assert(SP->session_status == invite_sent);
        cout << "\nVDB going active in onConnected(Client)" << endl;
	SP->session_status = active;
	SP->start_time = QTime::currentTime();
	SP->dest_port = (((sdp->session()).media()).front()).port();
	SP->dest_IP=(((sdp->session()).connection()).getAddress()).c_str();	
	SP->b=(((sdp->session()).bandwidths()).front()).kbPerSecond();//bandwidth in kbps; //bandwidth in bps
	SoftState.start_session(SP);

	SoftState.append(message);
	if(SP->pos_id==SoftState.Win->get_line())
		switch (SP->pos_id) {
			case 1:
				SoftState.Win->SlotLine1();
				break;
			case 2:
				SoftState.Win->SlotLine2();
				break;
			case 3:
				SoftState.Win->SlotLine3();
				break;
			case 4:
				SoftState.Win->SlotLine4();
				break;
			default:
				break;
		}
	else
		SoftState.Win->set_active(SP->pos_id);
}

void
UserAgentSip::onConnected(InviteSessionHandle h, const SipMessage& msg)
{
printf("\nVDB--> UserAgentSip::onConnected(InviteSessionHandle, const SipMessage& msg)\n");
	InfoLog( <<"onConnected - Server");
	cout << "\nVDB going active in onConnected()" << endl;
	SoftState.get_session((char*)h->getDialogId().data())->session_status = active;
	SoftState.get_session((char*)h->getDialogId().data())->start_time = QTime::currentTime();
	int code;
	char message[32];
	sscanf((char*)msg.brief().data(),"%*s %d %*s",&code);
	sprintf(message,"  %d OK - On Line",code);
	SoftState.append(message);
}

void
UserAgentSip::onStaleCallTimeout(ClientInviteSessionHandle)
{
	printf("\nVDB--> UserAgentSip::onStaleCallTimeout(ClientInviteSessionHandle)\n");
   	WarningLog(<< "onStaleCallTimeout");
}

void
UserAgentSip::onTerminated(InviteSessionHandle h, InviteSessionHandler::TerminatedReason reason, const SipMessage* msg)
{
	printf("\nVDB--> UserAgentSip::onTerminated(InviteSessionHandle h, InviteSessionHandler::TerminatedReason reason, const SipMessage* msg)\n");

	SessionPar* SP = SoftState.get_session((char*)h->getDialogId().data());
	if(SP==NULL) return;

	char parola[20];
	sprintf(parola,"status = %d",SP->session_status);
	
	cout << "\nVDB on terminted: " << parola << endl;
	int P=0;
	if(SP == NULL) {
		SoftState.append("  No call to close");
		return;
	}
	if(SP->session_status == active) {
		if(msg!=NULL)
			cout << "\nVDB --> " << msg->brief().data() << endl;
		P = SP->pos_id;
		SoftState.remove_session(SP->id);
//		CCThread remove_thread(&SoftState, RemoveFunction, SP);
//		remove_thread.run();
	}
	else if(SP->session_status == invite_received) {
		SP->session_status = cancelled;
		SoftState.append("  Remotely cancelled");
	}
	else if(SP->session_status == invite_sent) {
		SoftState.append("  Cancelling call");
		cout << "\nVDB --> Cancelling call" << endl;
		SoftState.stop_ringing(SP->pos_id);
		SoftState.remove_session(SP->id);
	}

	if (reason != InviteSessionHandler::PeerEnded){
		printf("\nVDB--> Call Terminated, local reason \n");
		char message[80];
		strcpy(message,"  ");
		strcat(message, (QTime::currentTime()).toString("hh:mm:ss"));
		strcat(message, ": Call locally terminated");
		SoftState.append(message);
		SoftState.append("  Off Line");
	}
	else {
		printf("\nVDB-->Call Terminated, remote reason \n");
		char message[80];
		strcpy(message,"  ");
		strcat(message, (QTime::currentTime()).toString("hh:mm:ss"));
		strcat(message, ": Call remotely terminated");
		SoftState.append(message);
		SoftState.append("  Off Line");
		SoftState.Win->set_inactive(P);
	}
}

void
UserAgentSip::onRedirected(ClientInviteSessionHandle, const SipMessage&)
{
	printf("\nVDB--> UserAgentSip::onRedirected(ClientInviteSessionHandle, const SipMessage& msg)\n");
}

void
UserAgentSip::onAnswer(InviteSessionHandle, const SipMessage&, const SdpContents&)
{
	printf("\nVDB--> UserAgentSip::onAnswer(InviteSessionHandle, const SipMessage& msg, const SdpContents&)\n");
}

void
UserAgentSip::onOffer(InviteSessionHandle, const SipMessage&, const SdpContents&)
//UserAgentSip::onOffer(InviteSessionHandle handle, const SipMessage& msg, const SdpContents& offer)
{
	printf("\nVDB--> UserAgentSip::onOffer(InviteSessionHandle handle, const SipMessage& msg, const SdpContents& offer)\n");
}

void
UserAgentSip::onOfferRequired(InviteSessionHandle, const SipMessage&)
{
	printf("\nVDB--> UserAgentSip::onOfferRequired(InviteSessionHandle, const SipMessage& msg)\n");
}

void
UserAgentSip::onOfferRejected(InviteSessionHandle, const SipMessage&)
{
	printf("\nVDB--> UserAgentSip::onOfferRejected(InviteSessionHandle, const SipMessage& msg)\n");
}

void
UserAgentSip::onDialogModified(InviteSessionHandle, InviteSession::OfferAnswerType, const SipMessage&)
{
	printf("\nVDB--> UserAgentSip::onDialogModified(InviteSessionHandle, InviteSession::OfferAnswerType oat, const SipMessage& msg)\n");
}

void
UserAgentSip::onInfo(InviteSessionHandle, const SipMessage&)
{
	printf("\nVDB--> UserAgentSip::onInfo(InviteSessionHandle, const SipMessage& msg)\n");
}

void
UserAgentSip::onInfoSuccess(InviteSessionHandle, const SipMessage&)
{
	printf("\nVDB--> UserAgentSip::onInfoSuccess(InviteSessionHandle, const SipMessage& msg)\n");
}

void
UserAgentSip::onInfoFailure(InviteSessionHandle, const SipMessage&)
{
	printf("\nVDB--> UserAgentSip::onInfoFailure(InviteSessionHandle, const SipMessage& msg)\n");
}

void
UserAgentSip::onRefer(InviteSessionHandle, ServerSubscriptionHandle, const SipMessage&)
{
	printf("\nVDB--> UserAgentSip::onRefer(InviteSessionHandle, ServerSubscriptionHandle, const SipMessage& msg)\n");
}

void
UserAgentSip::onReferAccepted(InviteSessionHandle, ClientSubscriptionHandle, const SipMessage&)
{
	printf("\nVDB--> UserAgentSip::onReferAccepted(InviteSessionHandle, ClientSubscriptionHandle, const SipMessage& msg)\n");
}

void
UserAgentSip::onReferRejected(InviteSessionHandle, const SipMessage&)
{
	printf("\nVDB--> UserAgentSip::onReferRejected(InviteSessionHandle, const SipMessage& msg)\n");
}


////////////////////////////////////////////////////////////////////////////////
// Registration Handler ////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
void 
UserAgentSip::onSuccess(ClientRegistrationHandle h, const SipMessage& response)
{ 
printf("\nVDB--> UserAgentSip::onSuccess(ClientRegistrationHandle h, const SipMessage& response)\n");
        assert(h.isValid());
        WarningLog(<< ": ClientRegistration-onSuccess - " + response.brief() );
	mCr = h->getHandle();
	//SoftState.append((char*)response.brief().data());
	if(mRegisterDuration)
		SoftState.registered = true;
	else
		SoftState.registered = false;
}

void
UserAgentSip::onFailure(ClientRegistrationHandle, const SipMessage& response)
{
	printf("\nVDB--> UserAgentSip::onFailure(ClientRegistrationHandle h, const SipMessage& response)\n");
	WarningLog(<< ": ClientRegistration-onFailure - " + response.brief() );
	SoftState.registered = false;
}

void
UserAgentSip::onRemoved(ClientRegistrationHandle)
{
	printf("\nVDB--> UserAgentSip::onRemoved(ClientRegistrationHandle h)\n");
	SoftState.registered = false;
}

int 
UserAgentSip::onRequestRetry(ClientRegistrationHandle, int, const SipMessage&)
//UserAgentSip::onRequestRetry(ClientRegistrationHandle h, int retryMinimum, const SipMessage& msg)
{
	printf("\nVDB--> UserAgentSip::onRequestRetry(ClientRegistrationHandle h, int retryMinimum, const SipMessage& msg)\n");
	return -1;
}

///Not Used
////////////////////////////////////////////////////////////////////////////////
// OutOfDialogHandler //////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
void 
UserAgentSip::onSuccess(ClientOutOfDialogReqHandle, const SipMessage& response)
{
printf("\nVDB--> UserAgentSip::onSuccess(ClientOutOfDialogReqHandle, const SipMessage& response)\n");
   InfoLog(<< response.header(h_CSeq).method() << "::OK: " << response );
}

void 
UserAgentSip::onFailure(ClientOutOfDialogReqHandle, const SipMessage& response)
{
printf("\nVDB--> UserAgentSip::onFailure(ClientOutOfDialogReqHandle, const SipMessage& response)\n");
   ErrLog(<< response.header(h_CSeq).method() << "::failure: " << response );
   if (response.exists(h_Warnings)) ErrLog  (<< response.header(h_Warnings).front());
	
	//SoftState.append((char*)response.brief().data());

}

void 
UserAgentSip::onReceivedRequest(ServerOutOfDialogReqHandle, const SipMessage&)
//UserAgentSip::onReceivedRequest(ServerOutOfDialogReqHandle, const SipMessage& request)
{
	printf("\nVDB--> UserAgentSip::onReceivedRequest(ServerOutOfDialogReqHandle, const SipMessage& request)\n");
}

void
UserAgentSip::onForkDestroyed(ClientInviteSessionHandle)
{
	printf("\nVDB--> UserAgentSip::onForkDestroyed(ClientInviteSessionHandle)\n");
}

void*
RemoveFunction(void* ptr)
{
	SessionPar* SP = (SessionPar*) ptr;
	cout << "\nVDB --> removing session on line " << SP->id << endl;
	SoftState.remove_session(SP->id);
	return NULL;
}

void* 
WhileFunction(void *ptr)
{
	session_data* ptr2 = (session_data*) ptr;
	SessionPar* SP = ptr2->SP;
	UserAgentSip* UAS= ptr2->UAS;
	cout << "\n VDB UAS = " << UAS << endl; 
	const char* mPtime = *(ptr2->mPtime);

	QTime ringtimer;
	ringtimer.start();
	int P = SP->pos_id;
	int S=0;
	while(!(ringtimer.elapsed() > 25000) && !(SP->session_status == active) && !(SP->session_status == rejected) && !(SP->session_status == cancelled)) {
		usleep(200000);
	}
	if(SP->session_status == active) {
		S=active;
		(*(SP->mSis))->provideAnswer(UAS->setSdp(mPtime,SP->source_port));
		(*(SP->mSis))->accept();
		SoftState.start_session(SP);
	} else if(SP->session_status == rejected) { //locally rejected
		S=rejected;
		(*(SP->mSis))->reject(486);
		SoftState.remove_session(SP->id);
		//(*(SP->mSis))->end();
	} else if(SP->session_status == cancelled) { //remotely cancelled
		S=cancelled;
		SoftState.remove_session(SP->id);
		//(*(SP->mSis))->end();
	}
	printf("\nVDB --> Line %d stop ringing (status = %d)\n", P, S);
	SoftState.stop_ringing(SP->pos_id);
	if(SP->session_status == invite_received) {
		(*(SP->mSis))->reject(603);
		SoftState.remove_session(SP->id);
		//(*(SP->mSis))->end();
	}
	return NULL;
}
